/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  June 3rd 2003 (April 2004)

  *****************************************************************

  Main.cpp

  Contains the main function for this solver.  Calls the appropriate
  parser then solves the DAG.  If the solution is satisfiable, then
  the solution is verified.

  *****************************************************************/
  
#include "noclause.h"
#include "memmanager.h"
#include "stats.h"

#include "parser.h"
#include "solver.h"



/********************** Global Variables **************************/

 // Problem structures
MemManager *g_memblock = NULL;
dag g_nodes;
input_arguments g_args;
stat_counters g_stats;
parse_counters g_desc;


 // Variables used to solve the problem
int g_num_nodes = 0;
int g_num_tmp_nogoods = 0;
int g_num_permanent_nogoods = 0;
byte g_is_satisfiable = UNKNOWN;
byte g_is_verified = FALSE;





/******************* Local Function Prototypes ********************/

void initializeGlobalVariables();





/**************************** Main ********************************/

int main(int argc, char *argv[]) {

  int i, retval, isCNF= FALSE, isSexpr = FALSE;
  char *filename = NULL;


 // Initialize the global variables
  initializeGlobalVariables();


 // Parse the input arguments
  for (i=1; i<argc; i++) {
    if (argv[i][0] != '-') {
      filename = argv[i];
    } else if (argv[i][1] == 'v') {
      g_args.verbose = ((int) argv[i][2]) - (int)'0';
      if (g_args.verbose >= 1) printf("Setting verbose level %d\n", g_args.verbose);
    } else if (strcasecmp(argv[i], "-cnf") == 0) {
      isCNF = TRUE;
    } else if (strcasecmp(argv[i], "-sexpr") == 0) {
      isSexpr = TRUE;
    } else if (strcasecmp(argv[i], "-nodc") == 0) {
      g_args.dcpropagation = FALSE;
    } else if (strcasecmp(argv[i], "-noreduce") == 0) {
      g_args.reductions = FALSE;
    } else if (strcasecmp(argv[i], "-noclean") == 0) {
      g_args.nogood_cleanup = FALSE;
    } else if (strcasecmp(argv[i], "-norestart") == 0) {
      g_args.restarts = FALSE;
    } else {
      printf("Unknown argument: %s\n", argv[i]);
      printf("Usage: %s <filename> [-v#] [-cnf|-sexpr] [-nodc] [-noreduce] [-norestart] [-noclean]\n\n", argv[0]);
      return -1;
    }
  }
  if (filename == NULL) {
    printf("Usage: %s <filename> [-v#] [-cnf|-sexpr] [-nodc] [-noreduce] [-norestart] [-noclean]\n\n", argv[0]);
    return -1;
  }
  if (isCNF && isSexpr) {
    printf("Input file cannot be in both CNF and S-Expression format\n");
    return -1;
  }



 // Allocate and fill the memory for the nodes (read the problem)
  if (isCNF)        retval = Parse_CNF(filename);
  else if (isSexpr) retval = Parse_Sexpr(filename);
  else              retval = Parse_ISCAS(filename);
  if (retval == NOCLAUSE_FILE_NOT_FOUND) {
    fprintf(stdout, "\nError: File \"%s\" not found\n\n", filename);
    return -2;
  } else if (retval != NOCLAUSE_OK) {
    fprintf(stdout, "\nAborting... Error occurred during parsing\n\n");
    return -3;
  }


 // Solve the expression
  Solver_SolveDAG();


 // Verify the solution
  if ((g_is_satisfiable) && (!isCNF) && (!isSexpr))
    Parse_VerifyISCAS(filename);


 // Print the statistics
  if (g_args.verbose == 0) {
    Stats_PrintLine(filename);
  } else {
    Stats_PrintResults();
    if      (g_is_satisfiable == TRUE)  printf("SAT\n");
    else if (g_is_satisfiable == FALSE) printf("UNSAT\n");
  }


 // Deallocate the tree
  delete g_atom_table;
  delete g_memblock;
  return 0;
}








/******************************************************************
  InitializeGlobalVariables
 ******************************************************************
 
  Description : Sets the default values of the global variables
  Parameters  : <none>
  Modifies    : Changes the value of variables in various structs
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
void initializeGlobalVariables() {

  g_args.verbose = 1;
  g_args.dcpropagation = TRUE;
  g_args.nogood_cleanup = TRUE;
  g_args.restarts = TRUE;
  g_args.reductions = TRUE;


 // Initialize the statistics
  Stats_Initialize();

}

