/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  June 3rd 2003 (April 2004)

  *****************************************************************

  NoClause.h

  Contains the declarations of globals, constants and macros used
  throughout the solver.

  Assumptions: FALSE must be 0 and TRUE must be 1

  *****************************************************************/
  
#ifndef __NOCLAUSE_H__
#define __NOCLAUSE_H__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// #define NDEBUG
#include <assert.h>

#include "vector.h"




/************************** Constants *****************************/

#define NOCLAUSE_OK                 1
#define NOCLAUSE_ERROR             -1
#define NOCLAUSE_INVALID_PARAMETER -2
#define NOCLAUSE_PARSE_ERROR       -3
#define NOCLAUSE_FILE_NOT_FOUND    -4
#define NOCLAUSE_EOF_ERROR         -5


#define FALSE                       0
#define TRUE                        1
#define DONT_CARE                   2
#define UNKNOWN                     3


#define CHOICE_REASON               0
#define CHILD_REASON                1
#define PARENT_REASON               2
#define NOGOOD_REASON               3
#define UIP_REASON                  4


#define ATOM_TYPE                   0
#define AND_TYPE                    1
#define IFF_TYPE                    2
#define NOT_TYPE                    3
#define OR_TYPE                     4
#define XOR_TYPE                    5
#define NUM_TYPES                   6

#define UNDEFINED_TYPE             10





/******************* Structure Declarations ***********************/


typedef char byte;


/*
   DAG record contains the information describing a problem.
   Each of the entries is an array of information such that
   we can get all the information for a node with the node's
   index.
    - pointer to the location in the static memblock
    - type of the node (AND, OR, NOT, IFF, XOR, ATOM)
    - value of the node (TRUE, FALSE, DONT_CARE, UNKNOWN)
    - decision level where the node was assigned its value
    - reason why the node was assigned
    - the node's VSIDS score (used when determining the next branch)
    - the index of the node in the ordered array as determined
      by the VSIDS score
    - list of nogoods that are being watched by the TRUE value
    - list of nogoods that are being watched by the FALSE value
    - list of And nodes that are being watched by this node
    - list of Or nodes that are being watched by this node
    - list of children which have this node as their DC watch parent
*/

typedef struct reasoninfo {
	int	type;
	union {
		int	*index;
		int	node;
	};
} reasoninfo;

typedef struct dag {

  int **loc;
  byte *type;
  byte *value;
  int *decision_level;
  reasoninfo *reason;

  int *score;
  int *score_index;

  Vector **true_watchlist;
  Vector **false_watchlist;
  Vector **and_watchlist;
  Vector **or_watchlist;
  Vector **dc_watchlist;

} dag;




/*
   Variables that are modified by the input arguments
    - verbosity of output
       0 - none (statistics printed in table format)
       1 - default (no solution output, summary output format)
       2 - minimal (updates on restarts, cleanups and # of decisions)
       3 - output all assignments and implications
       4 - also output conflict and backtrack information
       5 - add extensive parsing information
    - cleanup temporary NoGoods (default = TRUE)
    - use random restarts (default = TRUE)
    - propagate Don't Care values (default = TRUE)
    - reduce NoGoods by removing redundant literals (default = TRUE)
*/

typedef struct input_arguments {

  int verbose;
  int nogood_cleanup;
  int restarts;
  int dcpropagation;
  int reductions;

} input_arguments;





/*
   Statistic counters used in output summaries (see stats.cpp)
    - number of decisions chosen by the heuristic
    - number of choices avoided because of Don't Care values
    - implications inferred from the propagation stage
    - don't care implications inferred by DC propagations
    - number of nogoods learned during the solution
    - total number of literals in these nogoods
    - number of backtracks performed
    - total # of decision levels over which we backtracked
    - largest # of decision levels jumped by a single backtrack
    - number of nogoods examined in nogood reductions
    - total number of literals originally in these nogoods
    - number of removed redundant literals
    - number of random restarts
 */

typedef struct stat_counters {

  int num_choices;
  int num_avoided_choices;
  long long num_implications;
  long long num_dc_implications;
  int nogoods_learned;
  long long literals_learned;
  int num_backtracks;
  long long total_backtracks;
  int longest_backtrack;
  int num_nogoods_examined;
  long long num_literals_examined;
  long long num_redundant_literals;
  int num_restarts;

} stat_counters;




/*
   Various counters containing information on the problem description
    - number of atoms in the problem
    - number of nodes in the ORIGINAL problem (before merging common
      subformulas)
    - sum of the number of parents of every node
    - sum of the number of children in the problem
    - number of parents in the node with the most parents
    - most children for any node in the problem
    - number of AND nodes in the compact representation
    - number of OR nodes in the compact representation
    - number of NOT nodes in the compact representation
    - total number of children of AND nodes
    - total number of children of OR nodes
    - most children of an AND node in the problem
    - most children of an OR node in the problem
 */

typedef struct parse_counters {

  int num_atoms;
  int max_index;
  int num_total_parents;
  int num_total_children;
  int max_children;
  int max_parents;

  int num_and_nodes;
  int num_not_nodes;
  int num_or_nodes;
  int num_and_children;
  int num_or_children;
  int max_and_children;
  int max_or_children;

} parse_counters;





/********************** Global Variables **************************/


 // Global structures containing the problem and counters
extern MemManager *g_memblock;
extern dag g_nodes;
extern input_arguments g_args;
extern stat_counters g_stats;
extern parse_counters g_desc;


 // Global variables needed throughout the solving process
extern int g_num_nodes;
extern int g_num_tmp_nogoods;
extern int g_num_permanent_nogoods;
extern byte g_is_satisfiable;
extern byte g_is_verified;





/********************** Inline Functions **************************/

 // Swap the TRUE/FALSE value (undetermined for UNKNOWN/DONTCARE)
byte inline flipValue(byte value) {
  return value ^ 0x01;
}


 // Helper functions to access to contiguous block of memory
int inline getNumParents(int index) {
  return g_nodes.loc[index][1];
}
int inline getNumChildren(int index) {
  return g_nodes.loc[index][0];
}
int inline getChildNode(int index, int num) {
  return g_nodes.loc[index][2 + num];
}
int inline getParentNode(int index, int num) {
  return g_nodes.loc[index][2 + num + g_nodes.loc[index][0]];
}


 // Watch helper functions (watch children are the first in the list)
 // (last two entries contain the index of the node in the watch list)
void inline swapChildren(int index, int num1, int num2) {
  int tmp = g_nodes.loc[index][2 + num1];
  g_nodes.loc[index][2 + num1] = g_nodes.loc[index][2 + num2];
  g_nodes.loc[index][2 + num2] = tmp;
}
int inline getWatchLocation(int index, int num) {
  return g_nodes.loc[index][2 + num + g_nodes.loc[index][0] + g_nodes.loc[index][1]];
}
void inline setWatchLocation(int index, int num, int new_watch) {
  g_nodes.loc[index][2 + num + g_nodes.loc[index][0] + g_nodes.loc[index][1]] = new_watch;
}


 // Reason helper functions (a reason is an integer where the bottom
 // three bits are the type of reason
reasoninfo inline createNodeReason(byte type, int node) {
  reasoninfo	dst;
  if (type & 0xFFF8)
	  assert(NULL);
  dst.type = type;
  dst.node = node;
  return dst;
}
reasoninfo inline createNoGoodReason(byte type, int *nogood) {
  reasoninfo	dst;
  if (type & 0xFFF8)
	  assert(NULL);
  dst.type = type;
  dst.index = nogood;
  return (dst);
}
byte inline getReasonType(reasoninfo reason) {
  return (byte) (reason.type & 0x00000007);
}
int inline *getNoGoodReason(reasoninfo reason) {
  return (reason.index);
}
int inline getNodeReason(reasoninfo reason) {
  return (reason.node);
}



 // Debugging functions to print type and value of a node
char inline *printNodeType(int index) {
  switch (g_nodes.type[index]) {
    case ATOM_TYPE: return "ATOM";
    case AND_TYPE:  return "AND";
    case IFF_TYPE:  return "IFF";
    case NOT_TYPE:  return "NOT";
    case OR_TYPE:   return "OR";
    case XOR_TYPE:  return "XOR";
    default:        return "ERROR";
  }
}
char inline *printNodeValue(int index) {
  switch (g_nodes.value[index]) {
    case TRUE:      return "TRUE";
    case FALSE:     return "FALSE";
    case DONT_CARE: return "DON'T CARE";
    case UNKNOWN:   return "UNKNOWN";
    default:        return "ERROR";
  }
}




#endif
