/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  October 18th 2003 (April 2004)

  *****************************************************************

   NoGood.cpp

   This file contains the functions that can be used to create and
   store nogoods.  The nogoods are stored in a MemManager which can
   be cleaned up with NoGood_Cleanup.  The functions in this file
   are highly correlated with the undo functions as well as some
   propagate functions.

  *****************************************************************/

#include "solver.h"
#include "undo.h"

#include "memmanager.h"
#include "vector.h"
#include "uniquestack.h"



/********************** Global Variables **************************/

 // Database of NoGoods, UIPs and some additional memory for the
 // watch vectors (each node may watch a list of nogoods)
MemManager *g_db_large_nogoods = NULL;
MemManager *g_db_permanent_nogoods = NULL;
MemManager *g_db_uip_nogoods = NULL;
MemManager *g_watches = NULL;

 // Variables used when building and writing a nogood
dynamic_nogood g_dynamic;
int *g_nogood;
int g_nogood_size;
UniqueStack *g_removed_nodes;
bool g_wrote_nogood;





/******************* Local Function Prototypes ********************/

 // Modify the dynamic nogood
void analyzeConflict(int node, byte value, reasoninfo reason);
void insertNode(int node);
void removeNode(int node, byte value);

 // NoGood Reduction functions
void reduceNoGood();
void inline markRedundantNode(int node, byte value, int cause);
void markRedundantNeighbours(int node, int cause);

 // Write the dynamic nogood to the database
void writeNoGood();
void writeUIP();

 // Cleanup Functions
void cleanUIPDatabase();
void cleanTemporaryDatabase();








/********************** Inlined Functions *************************/


/******************************************************************
  WriteConflict
 ******************************************************************
 
  Description : Choose between writeNoGood and writeUIP
  Parameters  : <none>
  Modifies    : Writes a nogood to one of the three databases
  Assumptions : The dynamic nogood is correctly initialized
  Returns     : <none>

 ******************************************************************/
void inline writeConflict() {

 // Make sure the nogood is a UIP
  if (g_dynamic.remaining[g_dynamic.current_level] != 1)
    return;


 // Write the nogood if it is the first UIP
  if ((g_dynamic.size > 1) && (!g_wrote_nogood)) {
    reduceNoGood();
    if (g_nogood_size > 1)
      writeNoGood();
  }


 // Write the UIP to the UIP Database
  writeUIP();

}







/******************************************************************
  AddWatches
 ******************************************************************
 
  Description : Set this entry in the nogood as a watch
  Parameters  : entry - address of the entry that will be a watch
  Modifies    : allocates memory for a watch pointer that is added
                to the node's watchlist
  Assumptions : entry is not NULL
  Returns     : <none>

 ******************************************************************/
void inline addWatches(int *nogood) {

 // Add the watch for the first node in the nogood
  int node = nogood[0] >> 1;
  if (nogood[0] & 0x00000001) {
    *(nogood-2) = g_nodes.true_watchlist[node]->add((int)nogood) | 0x80000000;
  } else {
    *(nogood-2) = g_nodes.false_watchlist[node]->add((int)nogood) | 0x80000000;
  }


 // Add the watch for the second node in the nogood
  node = nogood[1] >> 1;
  if (nogood[1] & 0x00000001) {
    *(nogood-1) = g_nodes.true_watchlist[node]->add((int)nogood) | 0x80000000;
  } else {
    *(nogood-1) = g_nodes.false_watchlist[node]->add((int)nogood) | 0x80000000;
  }

}









/******************************************************************
  FindWatches
 ******************************************************************
 
  Description : Finds the nodes that have the highest decision
                levels and set them as the watches for this nogood
  Parameters  : nogood - pointer to the start of the nogood
  Modifies    : adds 2 watches to the nogood
  Assumptions : nogood is non-NULL and terminated by a 0
  Returns     : <none>

 ******************************************************************/
void inline findWatches(int *nogood) {

  int node, pos = 0, first_watch = -1, second_watch = -1;
  short last_decision = 0, second_last_decision = 0;


 // Find the location of both watches
  while (nogood[pos] != 0) {
    node = nogood[pos] >> 1;
    if (g_nodes.decision_level[node] > last_decision) {
      second_last_decision = last_decision;
      second_watch = first_watch;
      last_decision = g_nodes.decision_level[node];
      first_watch = pos;
    } else if (g_nodes.decision_level[node] > second_last_decision) {
      second_last_decision = g_nodes.decision_level[node];
      second_watch = pos;
    }
    pos++;
  }
 // Debugging <- prevented!
  assert( (first_watch != -1) && (second_watch != -1));



 // Move the watches and add them to the watch lists
  pos = nogood[first_watch];
  nogood[first_watch] = nogood[0];
  nogood[0] = pos;
  if (second_watch == 0)
    second_watch = first_watch;

  pos = nogood[second_watch];
  nogood[second_watch] = nogood[1];
  nogood[1] = pos;

}
















/***************** Externally Available Functions *****************/




/******************************************************************
  Initialize
 ******************************************************************
 
  Description : Allocates memory for the structures that will
                contain the dynamic nogood as well as the nogood
                database
  Parameters  : <none>
  Modifies    : Allocates all the memory required
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
void NoGood_Initialize() {


 // Allocate memory for dynamic nogood
  g_nogood = (int *) g_memblock->calloc (g_num_nodes, sizeof(int));
  g_nogood_size = 0;
  g_removed_nodes = new UniqueStack(g_num_nodes);
  g_dynamic.capacity = g_num_nodes;
  g_dynamic.nogood = (int *) g_memblock->calloc (g_dynamic.capacity, sizeof(int));
  g_dynamic.marks = (bool *) g_memblock->calloc (g_dynamic.capacity, sizeof(bool));
  g_dynamic.locs = (int *) g_memblock->calloc (g_dynamic.capacity, sizeof(int));
  g_dynamic.remaining = (short *) g_memblock->calloc (g_dynamic.capacity, sizeof(short));
  g_dynamic.size = 0;

  assert(g_nogood);
  assert(g_removed_nodes);
  assert(g_dynamic.nogood);
  assert(g_dynamic.marks);
  assert(g_dynamic.locs);
  assert(g_dynamic.remaining);


 // Allocate the database of permanent nogoods (and watches)
  g_db_large_nogoods = new MemManager(NOGOOD_DATABASE_SIZE);
  g_db_permanent_nogoods = new MemManager(NOGOOD_DATABASE_SIZE);
  g_db_uip_nogoods = new MemManager(NOGOOD_DATABASE_SIZE);
  g_watches = new MemManager(g_num_nodes * VECTOR_INITIAL_SIZE * sizeof(int*));
  assert(g_db_large_nogoods);
  assert(g_db_permanent_nogoods);
  assert(g_db_uip_nogoods);
  assert(g_watches);


 // Allocate and initialize the watch vectors for each node
  for (int i=0; i<g_num_nodes; i++) {
    g_nodes.true_watchlist[i] = new Vector(g_watches, VECTOR_INITIAL_SIZE);
    g_nodes.false_watchlist[i] = new Vector(g_watches, VECTOR_INITIAL_SIZE);
    g_dynamic.marks[i] = FALSE;
    g_dynamic.remaining[i] = 0;
  }


}







/******************************************************************
  Free
 ******************************************************************
 
  Description : Deallocate all the memory for the structures
  Parameters  : <none>
  Modifies    : deallocates memory
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
void NoGood_Free() {

  delete g_db_large_nogoods;
  delete g_db_permanent_nogoods;
  delete g_db_uip_nogoods;
  delete g_watches;

  for (int i=0; i<g_num_nodes; i++) {
    delete g_nodes.true_watchlist[i];
    delete g_nodes.false_watchlist[i];
    delete g_nodes.and_watchlist[i];
    delete g_nodes.or_watchlist[i];
    delete g_nodes.dc_watchlist[i];
  }

  delete g_removed_nodes;

}









/******************************************************************
  CreateNoGood
 ******************************************************************
 
  Description : Builds the initial nogood from a conflict node
  Parameters  : node   - the index to the node creating the conflict
                reason - the reason for the other assignment to the
                         node (conflict reason)
                continuing_backtrack - determine whether we should
                         write a new nogood or just UIPs since the
                         backtrack is continuing
  Modifies    : sets the dynamic nogood to the conflict reason
  Assumptions : everything has been allocated and g_dynamic has
                been initialized (all marks are set to FALSE)
  Returns     : <none>

 ******************************************************************/
void NoGood_CreateNoGood(int node, reasoninfo reason, bool continuing_backtrack) {


 // Create the initial conflict UIP
  g_wrote_nogood = continuing_backtrack;
  g_dynamic.current_level = g_decision_level;

  analyzeConflict(node, g_nodes.value[node], g_nodes.reason[node]);
  analyzeConflict(node, flipValue(g_nodes.value[node]), reason);
  writeConflict();


 // Print the initial UIP built
  if (g_args.verbose >= 4) {
    printf("Initial UIP is : ");
    for (int i=0; i<g_dynamic.size; i++)
      printf("%d%s ", g_dynamic.nogood[i]>>1,
        (g_nodes.decision_level[g_dynamic.nogood[i]>>1]==g_decision_level)?"*":"");
    printf("\n");
  }

}







/******************************************************************
  ExpandNoGood
 ******************************************************************
 
  Description : If the node is in the nogood, then replace it with
                the node's reason, expanding the nogood
  Parameters  : node - index of the node
  Modifies    : Adds information to the dynamic nogood, may write
                a nogood or UIP to the databases
  Assumptions : The dynamic nogood has been initialized and the
                initial conflict nogood is present (CreateNoGood)
  Returns     : <none>

 ******************************************************************/
void NoGood_ExpandUIP(int node) {


 // If we backtracked to another level, check if it is now a UIP
  if (g_decision_level < g_dynamic.current_level) {
    g_dynamic.current_level = g_decision_level;
    writeConflict();
  }



 // If the node is in the current nogood/UIP, then expand it
  if ((g_dynamic.marks[node] == TRUE) &&
      (getReasonType(g_nodes.reason[node]) != CHOICE_REASON)) {


   // Expand the conflict and remove the expanded node
    removeNode(node, g_nodes.value[node]);
    analyzeConflict(node, g_nodes.value[node], g_nodes.reason[node]);
    writeConflict();


   // Output the new nogood
    if (g_args.verbose >= 4) {
      printf("UIP is now (after expanding %d) (%d): ", node, g_dynamic.remaining[g_decision_level]);
      for (int i=0; i<g_dynamic.size; i++)
        printf("%d%s ", g_dynamic.nogood[i]>>1,
          (g_nodes.decision_level[g_dynamic.nogood[i]>>1]==g_decision_level)?"*":"");
      printf("\n");
    }


  }

}








/******************************************************************
  NoGood_ResetNoGood
 ******************************************************************
 
  Description : Empties the nogood efficiently
  Parameters  : <none>
  Modifies    : Resets all the variables in the dynamic nogood
  Assumptions : The nogood is initialized and has been built
  Returns     : <none>

 ******************************************************************/
void NoGood_ResetNoGood() {

  int i;


  if (g_dynamic.size > g_dynamic.current_level+1) {

    for (i=0; i<g_dynamic.current_level+1; i++) {
      g_dynamic.marks[g_dynamic.nogood[i] >> 1] = FALSE;
      g_dynamic.remaining[i] = 0;
    }
    for (i=g_dynamic.current_level+1; i<g_dynamic.size; i++) {
      g_dynamic.marks[g_dynamic.nogood[i] >> 1] = FALSE;
    }

  } else {

    for (i=0; i<g_dynamic.size; i++) {
      g_dynamic.marks[g_dynamic.nogood[i] >> 1] = FALSE;
      g_dynamic.remaining[i] = 0;
    }
    for (i=g_dynamic.size; i<g_dynamic.current_level+1; i++) {
      g_dynamic.remaining[i] = 0;
    }

  }

  g_dynamic.size = 0;

}


















/********************** Local Functions ***************************/





/******************************************************************
  AnalyzeConflict
 ******************************************************************
 
  Description : Expands the reason for a node into a nogood and
                appends it to the dynamic nogood
  Parameters  : node   - the node that is being expanded
                value  - the truth value the node was assigned
                         (since the node may be the conflict in
                         which case g_nodes.value[] is incorrect)
                reason - the reason for the assignment (again, for
                         the conflict, g_nodes.reason[] may be wrong)
  Modifies    : Adds the reason to the nogood
  Assumptions : Node 0 with value FALSE is never placed in the
                nogood
  Returns     : <none>

 ******************************************************************/
void analyzeConflict(int node, byte value, reasoninfo reason) {

  int i;
  byte type = getReasonType(reason);

 // Debugging information
  if (g_args.verbose >= 4) {
    printf("Processing %s node %d with value %s and %d children\n\tC_Reason = (%d,%p)\n",
           printNodeType(node), node, (value==TRUE)?"TRUE":
          (value==FALSE)?"FALSE":(value==UNKNOWN)?"UNKNOWN":
          "DONT_CARE", getNumChildren(node),
	  type,(int *)((int)getNoGoodReason(reason)*4));
    for (i=0; i<getNumChildren(node); i++)
      printf(" Child %d has value %s\n", getChildNode(node,i), printNodeValue(getChildNode(node,i)));
  }


  if ((type == 0) && (getNoGoodReason(reason) != NULL))
	  abort();

 // Determine the reason for the node's assignment
  if (type == CHOICE_REASON) {
    insertNode(node);


 // Combine the current nogood with the nogood reason
  } else if ((type == NOGOOD_REASON) || (type == UIP_REASON)) {

    int *reason_nogood = getNoGoodReason(reason);
    int pos = 0, index = -1;
    if (g_args.verbose >= 4) {
      if (type == NOGOOD_REASON)
        printf(" -> NoGood Reason :  ");
      else
        printf(" -> UIP Reason :  ");
      while (reason_nogood[pos] != 0) {
        printf("%d ", reason_nogood[pos] >> 1);
        pos++;
      }
      printf("\n");
      pos = 0;
    }

    while (reason_nogood[pos] != 0) {

      index = reason_nogood[pos]>>1;
      if (index != node) {
        insertNode(index);
      }
      pos++;
    }



 // Expand the parent reason
  } else if (type == PARENT_REASON) {

    int child, parent = getNodeReason(reason);
    assert((g_nodes.value[parent] == FALSE) || (g_nodes.value[parent] == TRUE));
    if (g_args.verbose >= 4) printf(" -> Parent Reason : %s is %s\n", printNodeType(parent), printNodeValue(parent));
    switch(g_nodes.type[parent]) {

      case AND_TYPE:
        insertNode(parent);
        if (g_nodes.value[parent] == FALSE) {
          for (i=0; i<getNumChildren(parent); i++) {
            child = getChildNode(parent, i);
            if (child != node) {
              assert(g_nodes.value[child] == TRUE);
              insertNode(child);
            }
          }
        }
        break;


      case OR_TYPE:
        insertNode(parent);
        if (g_nodes.value[parent] == TRUE) {
          for (i=0; i<getNumChildren(parent); i++) {
            child = getChildNode(parent, i);
            if (child != node) {
              assert(g_nodes.value[child] == FALSE);
              insertNode(child);
            }
          }
        }
        break;


      case IFF_TYPE:
      case XOR_TYPE:
        insertNode(parent);
        for (i=0; i<getNumChildren(parent); i++) {
          child = getChildNode(parent, i);
          if (child != node)
            insertNode(child);
        }
        break;


      case NOT_TYPE:   insertNode(parent);    break;
      default:         assert(NULL);
    }



 // Expand the child reason
  } else if (type == CHILD_REASON) {

    int child = getNodeReason(reason);
    if (g_args.verbose >= 4) printf(" -> Child Reason : %s is %s\n", printNodeType(node), printNodeValue(node));
    switch(g_nodes.type[node]) {

      case AND_TYPE:
        if (value == FALSE) {
          insertNode(child);
        } else {
          for (i=0; i<getNumChildren(node); i++) {
            insertNode(getChildNode(node, i));
          }
        }
        break;

      case OR_TYPE:
        if (value == TRUE) {
          insertNode(child);
        } else {
          for (i=0; i<getNumChildren(node); i++) {
            insertNode(getChildNode(node, i));
          }
        }
        break;


      case IFF_TYPE:
      case XOR_TYPE:
        for (i=0; i<getNumChildren(node); i++) {
          insertNode(getChildNode(node, i));
        }
        break;


      case NOT_TYPE:   insertNode(child);     break;
      default:         assert(NULL);
    }

  } else { // Verify that no other conditions exist
    printf("Error: Unknown Node Type (%d)\n", type);
    assert(NULL);
  }
}

















/******************************************************************
  InsertNode
 ******************************************************************
 
  Description : This function takes the index of a node, turns it
                into a nogood entry and places it at the end of the
                current dynamic nogood (g_dynamic)
  Parameters  : index - the index of the node
  Modifies    : adds an entry to g_dynamic, if the node is not
                already present
  Assumptions : g_dynamic is allocated
  Returns     : <none>

 ******************************************************************/
void insertNode(int node) {

 // Add the node to the nogood
  if ((g_dynamic.marks[node] == FALSE) && (g_nodes.decision_level[node] != 1)) {

    int entry = (node << 1) | (g_nodes.value[node]?0x1:0x0);
    g_dynamic.marks[node] = TRUE;
    g_dynamic.locs[node] = g_dynamic.size;
    g_dynamic.nogood[g_dynamic.size++] = entry;
    g_dynamic.remaining[g_nodes.decision_level[node]]++;

  }


}










/******************************************************************
  RemoveNode
 ******************************************************************
 
  Description : This function is used to remove a node from the
                nogood
  Parameters  : node  - the node to be removed
                value - the value of the node (make sure that the
                        literal is the one we want to remove)
  Modifies    : May remove a literal from the nogood
  Assumptions : The dynamic nogood is correctly initialized
  Returns     : <none>

 ******************************************************************/
void removeNode(int node, byte value) {

 // Remove the node by swapping it with the last node in the UIP
  if ((g_dynamic.marks[node] == TRUE) && (g_nodes.value[node] == value)) {

    int loc = g_dynamic.locs[node];
    g_dynamic.remaining[g_nodes.decision_level[node]]--;
    g_dynamic.marks[node] = FALSE;
    g_dynamic.nogood[loc] = g_dynamic.nogood[--g_dynamic.size];
    g_dynamic.locs[ g_dynamic.nogood[loc]>>1 ] = loc;

  }

}















/******************************************************************
  WriteNoGood
 ******************************************************************
 
  Description : Write the temporary nogood (in g_dynamic) to the
                nogood database
  Parameters  : <none>
  Modifies    : Allocates memory from the memblock and may create
                watches
  Assumptions : the nogood has at least 2 literals
  Returns     : <none>

 ******************************************************************/
void writeNoGood() {

 // Write the nogood to the database
  assert(g_nogood_size > 1);
  int *nogood;
  if (g_nogood_size < NOGOOD_MAX_KEEP_SIZE) {
    nogood = (int*) g_db_permanent_nogoods->calloc (g_nogood_size+3, sizeof(int));
    g_num_permanent_nogoods++;
  } else {
    nogood = (int*) g_db_large_nogoods->calloc (g_nogood_size+3, sizeof(int));
    g_num_tmp_nogoods++;
  }
  memcpy(&nogood[2], g_nogood, g_nogood_size * sizeof(int));
  nogood[g_nogood_size+2] = 0;

 // Add watches to the nogood
  findWatches(&nogood[2]);
  addWatches(&nogood[2]);

 // Update the variable scores for future branch decisions
  Solver_UpdateVariableScores(&nogood[2], g_nogood_size);

 // Update the nogood counters
  g_wrote_nogood = TRUE;
  g_stats.nogoods_learned++;
  g_stats.literals_learned += g_nogood_size;

 // Print the nogood for debugging
  if (g_args.verbose >= 3) {
    printf("Writing NoGood to db: ");
    for (int i=0; i<g_nogood_size+3; i++)
      printf("%d ", nogood[i]);
    printf("\n");
  }

}











/******************************************************************
  WriteUIP
 ******************************************************************
 
  Description : Write a UIP to the UIP database (no watches)
  Parameters  : <none>
  Modifies    : Allocates memory from the database and writes the
                current dynamic nogood as a UIP
  Assumptions : The databases are initialized
  Returns     : <none>

 ******************************************************************/
void writeUIP() {

  int i, *uip, uip_assignment = 0, level, largest = 0;


 // Write the UIP to memory
  uip = (int*) g_db_uip_nogoods->calloc (g_dynamic.size+1, sizeof(int));
  uip[g_dynamic.size] = 0;


 // Get the UIP assignment and next largest decision
  for (i=0; i<g_dynamic.size; i++) {
    uip[i] = g_dynamic.nogood[i];
    level = g_nodes.decision_level[uip[i] >> 1];
    if (level == g_dynamic.current_level) {
      uip_assignment = uip[i];
    } else if (level > largest) {
      largest = level;
    }
  }


 // Print the UIP for debugging
  if (g_args.verbose >= 3) {
    printf("Writing UIP to db: ");
    for (i=0; i<g_dynamic.size+1; i++)
      printf("%d ", uip[i]);
    printf("\n");
  }


 // Assign the UIP's flip value
  assert(uip_assignment > 0);
  if (uip_assignment & 0x1)    uip_assignment &= 0xFFFFFFFE;
  else                         uip_assignment |= 0x00000001;
  Undo_AddUIPAssignment(uip_assignment, uip, largest);

}














/******************************************************************
  ReduceNoGood
 ******************************************************************
 
  Description : Remove redundant literals from the nogood.  A lit is
                redundant if it would be propagated from the other
                literals in the nogood anyway. (Experimental results
                show that this reduces the size of the nogood by
                14% on average)
  Parameters  : <none>
  Modifies    : Writes a new copy of the nogood without the
                redundant literals
  Assumptions : The dynamic nogood is initialized, and the tmp
                nogood is allocated.
  Returns     : <none>

 ******************************************************************/
void reduceNoGood() {

  int i, node;

 //printf("Nogood is initially: ");
 //g_dynamic.nogood[g_dynamic.size] = 0;
 //NoGood_Print(g_dynamic.nogood);

 // Mark all the redundant literals
  if ((g_args.reductions == TRUE) && (g_dynamic.size < NOGOOD_MAX_REDUCE_SIZE)) {
    g_stats.num_nogoods_examined++;
    g_stats.num_literals_examined += g_dynamic.size;
    for (i=0; i<g_dynamic.size; i++) {
      node = g_dynamic.nogood[i] >> 1;
      if (!g_removed_nodes->member(node)) {
        markRedundantNeighbours(node, node);
      }
    }
  }


 // Copy the nogood but ignoring the redundant literals
  g_nogood_size = 0;
  for (i=0; i<g_dynamic.size; i++) {
    node = g_dynamic.nogood[i] >> 1;
    if (!g_removed_nodes->member(node))
      g_nogood[g_nogood_size++] = g_dynamic.nogood[i];
    else
      g_stats.num_redundant_literals++;
  }
  g_removed_nodes->empty();


 //printf("Nogood is now: ");
 //g_dynamic.nogood[g_dynamic.size] = 0;
 //NoGood_Print(g_dynamic.nogood);

}








/******************************************************************
  MarkRedundantNode
 ******************************************************************
 
  Description : This function is used to remove a node from the
                nogood
  Parameters  : node  - the node to be marked
                value - the value of the node to be marked (together
                        the two parameters form the literal)
                cause - the reason that forced this literal to be
                        redundant
  Modifies    : May add a literal to the removed stack
  Assumptions : The structures are correctly initialized
  Returns     : <none>

 ******************************************************************/
void inline markRedundantNode(int node, byte value, int cause) {

 // If the node is not the original cause (prevents us from marking
 // a literal redundant because of itself), in the nogood and not
 // already marked, then mark it and remove any additonal nodes that
 // make it redundant
  if ((node != cause) && (g_dynamic.marks[node] == TRUE) &&
      (!g_removed_nodes->member(node)) && (g_nodes.value[node] == value)) {
    g_removed_nodes->push(node);
    markRedundantNeighbours(node, cause);
  }

}







/******************************************************************
  MarkRedundantNeighbours
 ******************************************************************
 
  Description : Mark the parents/children that are made redundant
                because of a node in the current nogood
  Parameters  : node  - the node that is to be processed
                cause - the reason why any redundant nodes are to
                        be marked (this prevents cyclic dependencies
                        because the function is recursive)
  Modifies    : May add literals to g_removed_nodes through
                markRedundantNode()
  Assumptions : The structures are correctly initialized
  Returns     : <none>

 ******************************************************************/
void markRedundantNeighbours(int node, int cause) {

 // Mark the child of a NOT node
  if (g_nodes.type[node] == NOT_TYPE) {
    markRedundantNode(getChildNode(node, 0), flipValue(g_nodes.value[node]), cause);

 // Mark the children of a TRUE AND node
  } else if ((g_nodes.value[node] == TRUE) && (g_nodes.type[node] == AND_TYPE)) {
    for (int i=0; i<getNumChildren(node); i++)
      markRedundantNode(getChildNode(node, i), TRUE, cause);

 // Mark the children of a FALSE OR node
  } else if ((g_nodes.value[node] == FALSE) && (g_nodes.type[node] == OR_TYPE)) {
    for (int i=0; i<getNumChildren(node); i++)
      markRedundantNode(getChildNode(node, i), FALSE, cause);

  }


 // Mark the parents of the node
  for (int i=0; i<getNumParents(node); i++) {
    int parent = getParentNode(node, i);

   // Mark any NOT parents
    if (g_nodes.type[parent] == NOT_TYPE) {
      markRedundantNode(parent, flipValue(g_nodes.value[node]), cause);

   // Mark any AND parents if the child is FALSE
    } else if ((g_nodes.value[node] == FALSE) && (g_nodes.type[parent] == AND_TYPE)) {
      markRedundantNode(parent, FALSE, cause);

   // Mark any OR parents if the child is TRUE
    } else if ((g_nodes.value[node] == TRUE) && (g_nodes.type[parent] == OR_TYPE)) {
      markRedundantNode(parent, TRUE, cause);

    }
  }

}




















/*************************** Cleanup Code *************************/



/******************************************************************
  Cleanup
 ******************************************************************
 
  Description : Removes all unnecessary nogoods
  Parameters  : <none>
  Modifies    : too much to describe....
  Assumptions : the nogood database is initialized correctly
  Returns     : <none>

 ******************************************************************/
void NoGood_Cleanup() {

 // Output a status start message
  if (g_args.verbose >= 2) {
    printf("Starting cleanup (%d large nogoods) (%d/%d/%d blocks)\n",
           g_num_tmp_nogoods, g_db_large_nogoods->getNumBlocks(),
           g_db_permanent_nogoods->getNumBlocks(),
           g_db_uip_nogoods->getNumBlocks());
  }


 // Mark all nogoods and UIPs that are currently "reasons"
  Undo_CleanupNoGoods();

 // Cleanup the temporary nogoods
  cleanUIPDatabase();
  if (g_args.nogood_cleanup == TRUE) {
    cleanTemporaryDatabase();
  }


 // Output a status finish message
  if (g_args.verbose >= 2) {
    printf("Done cleanup (%d large nogoods) (%d/%d/%d blocks)\n",
           g_num_tmp_nogoods, g_db_large_nogoods->getNumBlocks(),
           g_db_permanent_nogoods->getNumBlocks(),
           g_db_uip_nogoods->getNumBlocks());
  }

}









/******************************************************************
  CleanUIPDatabase
 ******************************************************************
 
  Description : Removes all unnecessary uips
  Parameters  : <none>
  Modifies    : too much to describe....
  Assumptions : the nogood database is initialized correctly
  Returns     : <none>

 ******************************************************************/
void cleanUIPDatabase() {

  int *new_current, *old_current, *new_end, *old_end;
  int i, size, num_blocks, new_count;


 // Compress all the nogoods (keeping binary nogoods)
  new_count = 0;
  g_db_uip_nogoods->getBlock(new_count, &new_current, &new_end);
  new_end = g_db_uip_nogoods->getLastBlockEntry(new_count);
  num_blocks = g_db_uip_nogoods->getNumBlocks();


 // Look through all the blocks of UIPs
  for (i=0; i<num_blocks; i++) {

    g_db_uip_nogoods->getBlock(i, &old_current, &old_end);
    while (old_current < old_end-1) {

     // Get the size of this UIP
      assert(((*old_current) & 0x3FFFFFFF) == 0);
      size = 1;
      while ((old_current[size] & 0x3FFFFFFF) != 0)
        size++;


     // Keep marked UIPs
      if (old_current[0] != 0) {

       // If we are at the end of this block, get a new block
        if (new_end - new_current < size+1) {
          new_current[0] = 0;
          g_db_uip_nogoods->setNext(new_count++, new_current+1);
          g_db_uip_nogoods->getBlock(new_count, &new_current, &new_end);
          new_end = g_db_uip_nogoods->getLastBlockEntry(new_count);
        }


       // Write the UIP to memory
        memcpy(new_current, old_current, size * sizeof(int));


       // If it is in the UIP database, fix the pointer
        if (new_current[0] & 0x40000000) {

	  int	tmp = new_current[1];

//printf("Fixing up UIP entry 0x%x\n",tmp);
	  new_current[1] = g_uip_assignments->getstackentry(tmp);
//printf("\tnew_current[1] =  0x%x\n",new_current[1]);
	  g_uip_assignments->putstackentry(tmp,&(new_current[1]));
//printf("\tstack entry 0x%x now contains %p\n",tmp,&(new_current[1]));
        }


       // If it is the reason for an assignment, fix the pointer
        if (new_current[0] & 0x80000000) {
          int reason_loc = new_current[1];
	  new_current[1] = (int)g_nodes.reason[reason_loc].node;
	  g_nodes.reason[reason_loc] = createNoGoodReason(UIP_REASON, &(new_current[1]));

        }


       // Remove any marks
//assert(new_current[1] < g_num_nodes*2);
        new_current[0] = 0;
        new_current += size;

      }


      old_current += size;
    }

  }

 // Empty any block that is no longer used
  new_current[0] = 0;
  for (i=new_count+1; i<num_blocks; i++)
    g_db_uip_nogoods->setNext(i, NULL);
  g_db_uip_nogoods->setNext(new_count, new_current+1);

}









/******************************************************************
  CleanTemporaryDatabase
 ******************************************************************
 
  Description : Removes all unnecessary nogoods from the tmp db
  Parameters  : <none>
  Modifies    : Removes nogoods by compacting the kept nogoods
  Assumptions : The nogood database is initialized correctly and the
                nogoods that MUST be kept are also marked correctly
  Returns     : <none>

 ******************************************************************/
void cleanTemporaryDatabase() {

  MemManager *db = g_db_large_nogoods;
  int *new_current, *old_current, *new_end, *old_end;
  int i, size, pos, justified, num_blocks, new_count;
  int num_nogoods_examined = 0, old_num_nogoods = g_num_tmp_nogoods;


 // Initialize some data and get the first block
  g_num_tmp_nogoods = 0;
  new_count = 0;
  db->getBlock(new_count, &new_current, &new_end);
  new_end = db->getLastBlockEntry(new_count);
  num_blocks = db->getNumBlocks();


 // Look at each block containing nogoods
  for (i=0; i<num_blocks; i++) {

    db->getBlock(i, &old_current, &old_end);
    while (old_current < old_end-1) {

      assert(((*old_current) & 0x3FFFFFFF) == 0);
      assert((old_current[1] & 0x80000000) && (old_current[2] & 0x80000000));

     // Get the size of the nogood
      size = 4;
      justified = 0;
      while ((old_current[size] & 0x3FFFFFFF) != 0) {
        if (g_nodes.value[ old_current[size]>>1 ] == ((old_current[size]&0x1)?TRUE:FALSE))
          justified++;
        size++;
      }

      pos = 3;
      num_nogoods_examined++;
      assert(size-3 >= NOGOOD_MAX_KEEP_SIZE);


     // Keep the last 'x' nogoods or the relevant ones
      if ((old_current[0] != 0) || (old_num_nogoods - num_nogoods_examined < NOGOOD_RECENT_KEEP_COUNT) ||
          (size - justified < NOGOOD_IRRELEVANCE)) {


       // Get a new block if the nogood doesn't fit
        if (new_end - new_current < size+1) {
          new_current[0] = 0;
          db->setNext(new_count++, new_current+1);
          db->getBlock(new_count, &new_current, &new_end);
          new_end = db->getLastBlockEntry(new_count);
        }


       // Copy the nogood
        memcpy(new_current, old_current, size * sizeof(int));


       // Fix any pointers if the nogood is being used
        if (new_current[0] & 0x80000000) {
          int reason_loc = new_current[pos];
	  new_current[pos] = (int)g_nodes.reason[reason_loc].node;
	  g_nodes.reason[reason_loc] = createNoGoodReason(UIP_REASON, &(new_current[pos]));
        }
        assert(new_current[pos] < g_num_nodes*2);


       // Fix any watches
        g_num_tmp_nogoods++;
        NoGood_ChangeWatch(&new_current[3], &old_current[3]);

        new_current[0] = 0;
        new_current += size;


     // Remove the watches of any nogoods deleted
      } else {
        NoGood_RemoveWatch(&old_current[3], 0);
        NoGood_RemoveWatch(&old_current[3], 1);
      }

      old_current += size;
    }

  }


  new_current[0] = 0;
  for (i=new_count+1; i<num_blocks; i++)
    db->setNext(i, NULL);
  db->setNext(new_count, new_current+1);

}


