/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  October 18th 2003 (April 2004)

  *****************************************************************

   NoGood.h

   Contains definition of functions that are used to create and
   store nogoods.  Propagation of nogoods is performed in the file
   Propagate.cpp

  *****************************************************************/

#ifndef __NOGOOD_H__
#define __NOGOOD_H__

#include "noclause.h"



/************************** Constants *****************************/

#define NOGOOD_DATABASE_SIZE  /* 268435453 *//* 16777213 */ (0x100000 - 3)
#define NOGOOD_CLEANUP_COUNT        1000
#define NOGOOD_MAX_KEEP_SIZE         100
#define NOGOOD_MAX_REDUCE_SIZE       100

//#define NOGOOD_RECENT_KEEP_COUNT    (0.1 * NOGOOD_CLEANUP_COUNT)
#define NOGOOD_RECENT_KEEP_COUNT      25
#define NOGOOD_IRRELEVANCE            20



/******************* Structure Declarations ***********************/

/*
   Structure definition of a Dynamic NoGood that is used as we build
   a conflict during the backtracking stage.  Similar to a unique
   set, the dynamic nogood contains additional arrays in order to
   quickly determine which literals are present in the nogood and
   how many are at a given level.

    - Array of literals in the current nogood
    - Array of all possible literals (marked TRUE) if present in the
      current nogood
    - location of each literal in the current nogood (if present)
    - number of literals remaining at each level
    - current size of the nogood
    - maximum size of the nogood (= g_num_nodes)
    - current decision level that is being processed (to know if we
      backtracked over many decision levels without modifying the
      nogood)
 */

typedef struct dynamic_nogood {

  int *nogood;
  bool *marks;
  int *locs;
  short *remaining;
  int size, capacity;
  int current_level;

} dynamic_nogood;




/********************** Global Variables **************************/

 // Dynamic nogood and nogood databases defined for external use
 // (g_dynamic used in inline functions of this file and databases
 //  used in cleanup functions in Undo.cpp)
extern dynamic_nogood g_dynamic;
extern MemManager *g_db_large_nogoods;
extern MemManager *g_db_permanent_nogoods;



/********************** Inline Functions **************************/

 // Check if the choice is in the nogood (see Undo.cpp)
int inline NoGood_Member(int index) {
  return g_dynamic.marks[index];
}

 // Debugging function
void inline NoGood_Print(int *nogood) {
  int pos = 0;
  while (nogood[pos] != 0) {
    printf(" %d%s", nogood[pos] >> 1, (nogood[pos] & 0x1)?"T":"F");
    pos++;
  }
  printf("\n");
}


 // Remove the watches in a nogood (used in the cleanup functions)
void inline NoGood_RemoveWatch(int *nogood, int loc) {

  assert((loc == 0) || (loc == 1));
  int *vector_index_loc = (&nogood[loc]) - 2;
  int vector_index = (*vector_index_loc) & 0x7FFFFFFF;
  int swapped, *swapped_nogood;
  assert((*vector_index_loc) & 0x80000000);


 // Remove this watch in the corresponding watch vector
  if (nogood[loc] & 0x1) {
    assert( (int*)g_nodes.true_watchlist[ nogood[loc] >> 1 ]->getElementAt( vector_index ) == nogood );
    g_nodes.true_watchlist[ nogood[loc] >> 1 ]->remove( vector_index, &swapped );
  } else {
    assert( (int*)g_nodes.false_watchlist[ nogood[loc] >> 1 ]->getElementAt( vector_index ) == nogood );
    g_nodes.false_watchlist[ nogood[loc] >> 1 ]->remove( vector_index, &swapped );
  }


 // Removing involves a swap so make sure the swapped index is still consistent
  if (swapped != -1) {
    swapped_nogood = (int*) swapped;
    if (swapped_nogood[1] == nogood[loc]) {
      *(swapped_nogood-1) = vector_index | 0x80000000;
    } else {
      *(swapped_nogood-2) = vector_index | 0x80000000;
    }

  }

}



 // Move the watches in a nogood (used in Propagate.cpp when finding
 // new watches)
void inline NoGood_ChangeWatch(int *nogood, int *old_nogood) {

  int loc = (*(nogood - 2)) & 0x7FFFFFFF;
  if (nogood[0] & 0x1) {
    g_nodes.true_watchlist[ nogood[0] >> 1 ]->setElementAt( loc, (int)nogood );
  } else {
    g_nodes.false_watchlist[ nogood[0] >> 1 ]->setElementAt( loc, (int)nogood );
  }


  loc = (*(nogood - 1)) & 0x7FFFFFFF;
  if (nogood[1] & 0x1) {
    g_nodes.true_watchlist[ nogood[1] >> 1 ]->setElementAt( loc, (int)nogood );
  } else {
    g_nodes.false_watchlist[ nogood[1] >> 1 ]->setElementAt( loc, (int)nogood );
  }

}





/********************** Extern Functions **************************/

extern void NoGood_Initialize();
extern void NoGood_Free();
extern void NoGood_Cleanup();

extern void NoGood_CreateNoGood(int node, reasoninfo reason, bool cont);
extern void NoGood_ExpandUIP(int node);
extern void NoGood_ResetNoGood();



#endif
