/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  June 3rd 2003 (April 2004)

  *****************************************************************

  Parser.cpp

  Contains the functions required to parse an ISCAS problem written
  to a file into an internal DAG representation.  There have been
  many modifications and some functions should be replaced since
  they are now redundant.  The ISCAS parser was added last but has
  become the main parser.

  *****************************************************************/

#include "parser.h"
#include "stats.h"
#include "memmanager.h"
#include "uniquestack.h"



/********************** Global Variables **************************/

 // Temporary structures to hold the tmp representation of the problem
parse_node *g_parse_root = NULL;
parse_node *g_true_leaf, *g_false_leaf;
MemManager *g_tmp_memblock = NULL;

 // Hash tables used during parsing (should be replaced by the hash
 // table object defined in the header)
link_node *g_hashed_vars[PARSE_MAX_HASH_SPOTS];
link_node *g_hashed_iscas_nodes[PARSE_MAX_HASH_SPOTS];

 // Verification Table to keep the names of the atom signals
HashTable *g_atom_table;



/******************* Local Function Prototypes ********************/

 // Parsing Stage functions
void initializeProblemDescription();
int  buildParseTreeFromISCAS(char *filename);

 // Parsing helper functions
FILE inline *openInputFile(char *filename);
parse_node *addChild(parse_node *n, parse_node *child);
parse_node *createISCASAtom(char *s);
parse_node *createISCASNode(char *s);
parse_node *findISCASNode(char *s);

 // Merging helper functions
int mergeSubformulas();
unsigned int inline getHashCode(parse_node *n);
int compareNodeSet(parse_node *n, UniqueInfoStack *set, UniqueInfoStack *blank);
int isEquivalentNode(parse_node *n1, parse_node *n2);
int replaceNode(parse_node *replaced, parse_node *fixed, UniqueInfoStack *blank);
int replaceNotNode(parse_node *replaced, parse_node *fixed);
int findNode(void *source, link_node *list);


 // Compact representation helper functions
int compactRepresentation();
int indexNode(parse_node *n, int index, int *parent_watch, parse_node **node_list);
int convertNode(parse_node *n, int index, int *loc);


 // Node creation helper functions
parse_node *createAtom(char *id);
parse_node *createNode();
link_node *createLinkNode(parse_node *n);
void removeParentTwin(link_node *n);
void removeChildTwin(link_node *n);


 // Other parsing functions
int  buildParseTreeFromSexpr(char *filename);
int  buildParseTreeFromCNF(char *filename);
parse_node *buildNodeFromSexpr(FILE *infile);
parse_node *combineChild(parse_node *n, parse_node *child);
int getToken(FILE *infile, char *s);
int countNegativeChildren(parse_node *n);
void swapNegativeChildren(parse_node *n);
parse_node *moveChildren(parse_node *n, link_node *child_list);


 // Functions to add variables to a hash table
void createVariableHashTable();
void deleteVariableHashTable();
unsigned int getVariableLocation(char *s);











/***************** Externally Available Functions *****************/



/******************************************************************
  Parse_ISCAS
 ******************************************************************
 
  Description : Parse the ISCAS circuit contained in the file and
                create an internal DAG.
  Parameters  : filename - null-terminated string containing the
                           location of the file
  Modifies    : g_indexed_nodes is set to point to all the nodes
                that will be contained in g_memblock (the tree)
  Returns     : NOCLAUSE_OK if all went well
                the return value of the helper functions otherwise
  Note        : All parsing functions use the global g_parse_root
                variable

 ******************************************************************/
int Parse_ISCAS(char *filename) {

  int retval;

 // Initialize structures that will be needed for parsing
  initializeProblemDescription();
  createVariableHashTable();
  g_tmp_memblock = new MemManager(PARSE_DEFAULT_BLOCKSIZE);


 // Read in the parse tree from an input file
  if (g_args.verbose >= 2)  fprintf(stdout, "\nBuilding parse tree\n");
  Stats_StartTimer(STATS_PARSING);
  retval = buildParseTreeFromISCAS(filename);
  Stats_StopTimer(STATS_PARSING);
  if (retval != NOCLAUSE_OK)
    return retval;
  if (g_args.verbose >= 2) printf(" Found %d AND nodes, %d OR nodes, %d NOT nodes, %d ATOMS\n", g_desc.num_and_nodes, g_desc.num_or_nodes, g_desc.num_not_nodes, g_desc.num_atoms);


 // Find identical subformulas (build a DAG)
  if (g_args.verbose >= 2)  fprintf(stdout, "\nMerging common subformulas\n");
  Stats_StartTimer(STATS_MERGING);
  retval = mergeSubformulas();
  Stats_StopTimer(STATS_MERGING);
  if (retval != NOCLAUSE_OK)
    return retval;


 // Reduce the parse expression to a compact representation
  if (g_args.verbose >= 2)  fprintf(stdout, "\nConverting to compact representation (%d nodes)\n", g_num_nodes);
  Stats_StartTimer(STATS_COMPACT);
  retval = compactRepresentation();
  Stats_StopTimer(STATS_COMPACT);
  if (retval != NOCLAUSE_OK)
    return retval;


 // Cleanup and make sure the global root is correctly built
  deleteVariableHashTable();
  delete g_tmp_memblock;
  if (g_args.verbose >= 2) fprintf(stdout, "\nBuilding Complete\n");
  return NOCLAUSE_OK;
}







/******************************************************************
  Parse_VerifyISCAS
 ******************************************************************

  Description : Verify that a satisfiable solution determined for an
                ISCAS circuit is still correct in the original
                circuit.
  Parameters  : filename - the original circuit that was used to
                           build the problem by Parse_ISCAS()
  Modifies    : Assigns a TRUE/FALSE value to is_verified
  Assumptions : The Hash Table contains ALL the atoms from the
                original formula
  Returns     : NOCLAUSE_FILE_NOT_FOUND or NOCLAUSE_OK
  Note        : If the problem cannot be verified, then there is a
                logical error in the solving process (BIG PROBLEM!)

 ******************************************************************/
int Parse_VerifyISCAS(char *filename) {

  int token_type;
  char *token = NULL, identifier[PARSE_MAX_TOKLEN], out_id[PARSE_MAX_TOKLEN];
  byte output = FALSE, type = UNDEFINED_TYPE, value = UNKNOWN;
  HashTable verification_table(g_memblock, PARSE_MAX_HASH_SPOTS);
  ISCASTokenizer parser(filename);
  out_id[0] = '\0';

  if (g_args.verbose >= 2) printf("Verifying the solution\n");

 // Check that the file was opened and get the first token
  token_type = parser.nextToken(&token);
  if (token_type == NOCLAUSE_FILE_NOT_FOUND)
    return token_type;


 // Read all the tokens until the end of file
  while (token_type != PARSE_TOKEN_EOLN) {

   // Add an atom with the value in the solution to a new hash table
    if (token_type == PARSE_TOKEN_INPUT) {

      int index = g_atom_table->getInfo(token);
      assert(index != -1);
      byte atom_value = g_nodes.value[index];
      if (atom_value == DONT_CARE) atom_value = TRUE;
     // printf("Adding input %s (%d)\n", token, atom_value);
      verification_table.add(token, atom_value);
      token_type = parser.nextToken(&token);
      assert(token_type == PARSE_TOKEN_EOLN);

    } else {

      sprintf(identifier, "%s", token);
      if (token_type == PARSE_TOKEN_OUTPUT) {
        output = TRUE;
      }

     // Get the token and determine the value from the other gates
     // already added to the hash table
      token_type = parser.nextToken(&token);
      assert(token_type == PARSE_TOKEN_DEFAULT);
      if        (strcasecmp("AND", token) == 0) {  type = AND_TYPE;  value = TRUE;
      } else if (strcasecmp("OR" , token) == 0) {  type = OR_TYPE;   value = FALSE;
      } else if (strcasecmp("NOT", token) == 0) {  type = NOT_TYPE;  value = UNKNOWN;
      //} else if (strcasecmp("XOR", token) == 0) {  type = XOR_TYPE;
      //} else if (strcasecmp("IFF", token) == 0) {  type = IFF_TYPE;
      } else { printf("Unknown token \"%s\"\n", token); assert(NULL); }

     // Get all the children and determine the truth value of the gate
      token_type = parser.nextToken(&token);
      assert(token_type == PARSE_TOKEN_DEFAULT);
      while (token_type != PARSE_TOKEN_EOLN) {
        int child_value = verification_table.getInfo(token);
        assert(child_value != -1);
        if ((type == AND_TYPE) && (child_value == FALSE)) {
          value = FALSE;
          break;
        } else if ((type == OR_TYPE) && (child_value == TRUE)) {
          value = TRUE;
          break;
        } else if (type == NOT_TYPE) {
          value = flipValue(child_value);
          break;
        }
        token_type = parser.nextToken(&token);
      }
      while (token_type != PARSE_TOKEN_EOLN)
        token_type = parser.nextToken(&token);

    // Add the new gate to the hash table with the assigned truth value
    // printf("Adding Gate %s (%d)\n", identifier, value);
      verification_table.add(identifier, value);
      if (output == TRUE) {
        sprintf(out_id, "%s", identifier);
        output = FALSE;
      }

    }

    token_type = parser.nextToken(&token);
  }


 // Verify the output of the circuit
  assert(out_id[0] != '\0');
  value = verification_table.getInfo(out_id);
  if (value == TRUE) {
    if (g_args.verbose >= 2) printf("Solution Verified\n");
    g_is_verified = TRUE;
  } else {
    printf("Error: Unable to verify solution (%d)\n", value);
    g_is_verified = FALSE;
  }


  return NOCLAUSE_OK;

}








/******************************************************************
  Parse_Sexpr
 ******************************************************************
 
  Description : Parse the S-expression contained in the file and
                create an internal DAG.
  Parameters  : filename - null-terminated string containing the
                           location of the file
  Modifies    : g_indexed_nodes is set to point to all the nodes
                that will be contained in g_memblock (the tree)
  Returns     : NOCLAUSE_OK if all went well
                the return value of the helper functions otherwise
  Note        : All parsing functions use the global g_parse_root
                variable and must free the tree afterwards.
                This function must be fixed to account for watches.

 ******************************************************************/
int Parse_Sexpr(char *filename) {

  int retval;

 // This function has not been modified to take into account watches
 // for AND and OR nodes.  We need to keep g_desc.num_and_nodes and
 // g_desc.num_or_nodes updated as we build the DAG
  assert(NULL);
  initializeProblemDescription();


 // Initialize structures that will be needed for parsing
  createVariableHashTable();
  g_tmp_memblock = new MemManager(PARSE_DEFAULT_BLOCKSIZE);


 // Read in the parse tree from an input file
  if (g_args.verbose >= 2)  fprintf(stdout, "\nBuilding parse tree\n");
  Stats_StartTimer(STATS_PARSING);
  retval = buildParseTreeFromSexpr(filename);
  Stats_StopTimer(STATS_PARSING);
  if (retval != NOCLAUSE_OK)
    return retval;


 // Find identical subformulas (build a DAG)
  if (g_args.verbose >= 2)  fprintf(stdout, "\nMerging common subformulas\n");
  Stats_StartTimer(STATS_MERGING);
  retval = mergeSubformulas();
  Stats_StopTimer(STATS_MERGING);
  if (retval != NOCLAUSE_OK)
    return retval;


 // Reduce the parse expression to a compact representation
  if (g_args.verbose >= 2)  fprintf(stdout, "\nConverting to compact representation\n");
  Stats_StartTimer(STATS_COMPACT);
  retval = compactRepresentation();
  Stats_StopTimer(STATS_COMPACT);
  if (retval != NOCLAUSE_OK)
    return retval;


 // Cleanup and make sure the global root is correctly built
  deleteVariableHashTable();
  delete g_tmp_memblock;
  if (g_args.verbose >= 2) fprintf(stdout, "\nBuilding Complete\n");
  return NOCLAUSE_OK;
}










/******************************************************************
  Parse_CNF
 ******************************************************************
 
  Description : Read a CNF file in the standard DIMACS format and
                create an internal DAG.
  Parameters  : filename - null-terminated string containing the
                           location of the file
  Modifies    : g_indexed_nodes is set to point to all the nodes
                that will be contained in g_memblock (the tree)
  Returns     : NOCLAUSE_OK if all went well
                the return value of the helper functions otherwise
  Note        : This function has been fixed and should work ok.
                Furthermore, the function ignores any additional
                clauses (reads only the amount of clauses defined
                in the header)

 ******************************************************************/
int Parse_CNF(char *filename) {

  int retval;

 // Initialize structures that will be needed for parsing
  initializeProblemDescription();
  createVariableHashTable();
  g_tmp_memblock = new MemManager(PARSE_DEFAULT_BLOCKSIZE);


 // Read in the parse tree from an input file
  if (g_args.verbose >= 2)  fprintf(stdout, "\nBuilding parse tree\n");
  Stats_StartTimer(STATS_PARSING);
  retval = buildParseTreeFromCNF(filename);
  Stats_StopTimer(STATS_PARSING);
  if (retval != NOCLAUSE_OK)
    return retval;


 // Find identical subformulas (build a DAG)
 // This removes identical NOT nodes that were added during parsing
  if (g_args.verbose >= 2)  fprintf(stdout, "\nMerging common subformulas\n");
  Stats_StartTimer(STATS_MERGING);
  retval = mergeSubformulas();
  Stats_StopTimer(STATS_MERGING);
  if (retval != NOCLAUSE_OK)
    return retval;


 // Reduce the parse expression to a compact representation
  if (g_args.verbose >= 2)  fprintf(stdout, "\nConverting to compact representation\n");
  Stats_StartTimer(STATS_COMPACT);
  retval = compactRepresentation();
  Stats_StopTimer(STATS_COMPACT);
  if (retval != NOCLAUSE_OK)
    return retval;


 // Cleanup and make sure the global root is correctly built
  deleteVariableHashTable();
  delete g_tmp_memblock;
  if (g_args.verbose >= 2) fprintf(stdout, "\nBuilding Complete\n");
  return NOCLAUSE_OK;
}






























/*********************** Local Functions **************************/


/******************************************************************
  InitializeProblemDescription
 ******************************************************************
 
  Description : Initialize the parsing counters in g_desc
  Parameters  : <none>
  Modifies    : Assigns 0 to many description counters
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
void initializeProblemDescription() {

  g_desc.max_index = 0;
  g_desc.num_atoms = 0;
  g_desc.num_total_parents = 0;
  g_desc.num_total_children = 0;
  g_desc.max_parents = 0;
  g_desc.max_children = 0;

  g_desc.num_and_nodes = 0;
  g_desc.num_not_nodes = 0;
  g_desc.num_or_nodes = 0;

  g_desc.num_and_children = 0;
  g_desc.num_or_children = 0;
  g_desc.max_and_children = 0;
  g_desc.max_or_children = 0;

}








/******************************************************************
  BuildParseTreeFromISCAS
 ******************************************************************

  Description : Read an ISCAS file using an ISCASTokenizer and
                build the corresponding temporary DAG in parse_root
  Parameters  : filename - the file containing the DAG
  Modifies    : Memory is allocated for the DAG
  Assumptions : Problem description has been initialized
  Returns     : NOCLAUSE_FILE_NOT_FOUND or NOCLAUSE_OK

 ******************************************************************/
int buildParseTreeFromISCAS(char *filename) {

  int token_type;
  char *token = NULL;
  parse_node *n;
  ISCASTokenizer parser(filename);


 // Check that the file was opened and get the first token
  token_type = parser.nextToken(&token);
  if (token_type == NOCLAUSE_FILE_NOT_FOUND)
    return token_type;


 // Read all the tokens until the end of file
  while (token_type != PARSE_TOKEN_EOLN) {

    if (token_type == PARSE_TOKEN_INPUT) {
      n = createISCASAtom(token);
      token_type = parser.nextToken(&token);
      assert(token_type == PARSE_TOKEN_EOLN);
    } else {

     // Create a new node for the gate
      n = createISCASNode(token);
      if (token_type == PARSE_TOKEN_OUTPUT) {
        g_parse_root = n;
      }

     // Set the type of the gate
      token_type = parser.nextToken(&token);
      assert(token_type == PARSE_TOKEN_DEFAULT);
      if        (strcasecmp("AND", token) == 0) {  n->type = AND_TYPE; g_desc.num_and_nodes++;
      } else if (strcasecmp("OR" , token) == 0) {  n->type = OR_TYPE;  g_desc.num_or_nodes++;
      } else if (strcasecmp("NOT", token) == 0) {  n->type = NOT_TYPE; g_desc.num_not_nodes++;
      } else if (strcasecmp("XOR", token) == 0) {  n->type = XOR_TYPE;
      } else if (strcasecmp("IFF", token) == 0) {  n->type = IFF_TYPE;
      } else { printf("Unknown token \"%s\"\n", token); assert(NULL); }

     // Add the children to this node
      token_type = parser.nextToken(&token);
      assert(token_type == PARSE_TOKEN_DEFAULT);
      while (token_type != PARSE_TOKEN_EOLN) {
        parse_node *tmp = findISCASNode(token);
        addChild(n, tmp);
        token_type = parser.nextToken(&token);
      }

    }

    token_type = parser.nextToken(&token);
  }



  assert(g_parse_root);
  return NOCLAUSE_OK;

}








/******************************************************************
  OpenInputFile
 ******************************************************************
 
  Description : Open a file as Read-Only
  Parameters  : filename - null-terminated string containing the
                           location of the file
  Returns     : A non-Null pointer to the opened file

 ******************************************************************/
FILE inline *openInputFile(char *filename) {

  FILE *infile = NULL;
  
  assert(filename);
  infile = fopen(filename, "r");
  return infile;
}









/******************************************************************
  AddChild
 ******************************************************************

  Description : Add a child to a given parse_node
  Parameters  : n     - the create node in question
                child - an initialized parse_node to be added to n's
                        child list
  Modifies    : Adds child to the child list of n and increments the
                number of children (same for parent list of child)
  Assumptions : Both nodes are initialized
  Returns     : n after the changes have been made

 ******************************************************************/
parse_node *addChild(parse_node *n, parse_node *child) {

  link_node *childlink, *parentlink;

 /* Create the new child link nodes */
  childlink = createLinkNode(child);
  parentlink = createLinkNode(n);
  childlink->twin = parentlink;
  parentlink->twin = childlink;

 /* Add the links to the list of parents and children */
  if (n->children != NULL)
    n->children->prev = childlink;
  childlink->next = n->children;
  n->children = childlink;
  n->num_children++;

  if (child->parents != NULL)
    child->parents->prev = parentlink;
  parentlink->next = child->parents;
  child->parents = parentlink;
  child->num_parents++;

  return n;
}






/******************************************************************
  CreateISCASAtom
 ******************************************************************

  Description : Creates an atom for an ISCAS problem
  Parameters  : s - the string identifier for this atom
  Modifies    : Allocates memory for a new node
  Assumptions : Memory manager is initialized
  Returns     : The new parse_node representing the atom

 ******************************************************************/
parse_node *createISCASAtom(char *s) {
  return createAtom(s);
}






/******************************************************************
  CreateISCASNode
 ******************************************************************

  Description : Creates a new node for a gate in an ISCAS circuit
  Parameters  : s - the identifier representing the gate
  Modifies    : Allocates memory for a new parse_node and adds it
                to the hash table for future reference
  Assumptions : Memory manager and hash table are initialized
  Returns     : A newly create parse_node

 ******************************************************************/
parse_node *createISCASNode(char *s) {

  parse_node *n = createNode();
  link_node *link = createLinkNode(n);
  unsigned int loc = getVariableLocation(s);
  n->identifier = (char*) g_tmp_memblock->calloc(strlen(s)+1, sizeof(char));
  strcpy(n->identifier, s);

 // Store it in a hash table
  link->next = g_hashed_iscas_nodes[loc];
  g_hashed_iscas_nodes[loc] = link;

  return n;
}





/******************************************************************
  FindISCASNode
 ******************************************************************

  Description : Find the node in the ISCAS table represented by the
                given identifier string
  Parameters  : s - string that is the name of the identifier
  Modifies    : <none>
  Assumptions : Everything initialized and the atom is present in the
                hash table (otherwise the ISCAS file is incorrect)
  Returns     : The node that represents the ISCAS gate

 ******************************************************************/
parse_node *findISCASNode(char *s) {


 // Look for the node in the atom hash table
  unsigned int loc = getVariableLocation(s);
  link_node *list = g_hashed_vars[loc];
  while (list != NULL) {
    if (strcasecmp(s, ((parse_node*)(list->n))->identifier) == 0)
      return (parse_node*)(list->n);
    list = list->next;
  }


 // Look for the node in the ISCAS hash table
  list = g_hashed_iscas_nodes[loc];
  while (list != NULL) {
    if (strcasecmp(s, ((parse_node*)(list->n))->identifier) == 0)
      return (parse_node*)(list->n);
    list = list->next;
  }


  printf("Could not find node definition for \"%s\" in ISCAS file\n", s);
  assert(NULL);
  return NULL;
}


















/***************** Merging Subformula Functions *******************/



/******************************************************************
  MergeSubformulas
 ******************************************************************
 
  Description : Merge identical subtrees in the DAG (reducing the
                size of the representation
  Parameters  : <none>
  Modifies    : the DAG contained in root may have some nodes merged
                together (identical subtrees are joined)
  Returns     : NOCLAUSE_OK if all went well
                NOCLAUSE_ERROR otherwise

 ******************************************************************/
int mergeSubformulas() {

  int i, num_nodes = g_num_nodes;
  parse_node *next_node;
  link_node *current;
  UniqueInfoStack *set;
  UniqueInfoStack *blank;


 // Initialize a queue of marked identical nodes
  assert(g_parse_root);
  assert(g_desc.num_atoms>0);
  set = new UniqueInfoStack(g_desc.max_index);
  blank = new UniqueInfoStack(g_desc.max_index+1);


 // Perform a search starting at the nodes of the tree, marking
 // all the leaves
  if (g_args.verbose >= 4) fprintf(stdout, " Looking for identical atoms\n");
  for (i=0; i<PARSE_MAX_HASH_SPOTS; i++) {

    current = g_hashed_vars[i];
    while (current != NULL) {
      if (((parse_node*)(current->n))->num_parents > 1) {
        if (g_args.verbose >= 5) printf("  Adding node %s with %d parents to the queue\n", ((parse_node*)current->n)->identifier, ((parse_node*)current->n)->num_parents);
        set->push(((parse_node*)current->n)->index, current->n);
      }
      current = current->next;
    }
  }



 // Work through the marked nodes in the queue
  if (g_args.verbose >= 4) fprintf(stdout, " Comparing possibly identical nodes\n");
  while (!set->isEmpty()) {

   // Replaced nodes will have an index of -1
    do {
      set->pop((void**)&next_node);
    } while ((!set->isEmpty()) && (next_node->index < 0));


    compareNodeSet(next_node, set, blank);

  }


  delete blank;
  delete set;


  if (g_args.verbose >= 2) fprintf(stdout, " Merging removed %d nodes (%d remaining, %d are atoms)\n", (num_nodes-g_num_nodes), g_num_nodes, g_desc.num_atoms);
  return NOCLAUSE_OK;

}






/******************************************************************
  GetHashCode
 ******************************************************************

  Description : Combine the indices of all the children and assign
                hash code to the given node such that nodes of the
                same type with the same children will have the same
                hash code.
  Parameters  : n - the parse node to assign a code
  Modifies    : <none>
  Assumptions : n is initialized correctly with a type
  Returns     : A hash code integer representing the node

 ******************************************************************/
unsigned int inline getHashCode(parse_node *n) {

  link_node *current = n->children;
  unsigned long total = n->type;

 // Combine the children to the total
  while (current != NULL) {
    total += (unsigned long) current->n;
    current = current->next;
  }

 // Calculate the hash number
  return (unsigned int) (total % PARSE_MERGING_HASH_CODE);
}










/******************************************************************
  CompareNodeSet
 ******************************************************************

  Description : The parents of node n are compared and if any two
                parent pairs p1/p2 have the same children, then
                p1 is removed and p2 is added to the set of nodes
                to be examined.
  Parameters  : n     - Compare the parents of n to see if any have
                        the same parents
                set   - The set of nodes that may have similar
                        parents
                blank - A blank unique stack used to compare parent
                        nodes when replacing nodes that have 
                        identical children
  Modifies    : Adds nodes to the stack 'set'
  Assumptions : The unique stacks are initialized and blank is empty
  Returns     : SOLVER_OK

 ******************************************************************/
int compareNodeSet(parse_node *n, UniqueInfoStack *set, UniqueInfoStack *blank) {

  link_node *hash[PARSE_MERGING_HASH_CODE];
  link_node *current, *current_hash;
  parse_node *not_node = NULL;
  int i;
  unsigned int loc = 0;


 // Try to memcpy this?
  for (i=0; i<PARSE_MERGING_HASH_CODE; i++) {
    hash[i] = NULL;
  }



 // Compare each of n's parents using a Hash table
  current = n->parents;
  while (current != NULL) {


   // All NOT parents must be identical
    if ( ((parse_node*)current->n)->type == NOT_TYPE) {
      if (not_node == NULL) {
        not_node = (parse_node*)current->n;
      } else {
       // This can be faster sometimes... (for s-expr merging)
        replaceNotNode((parse_node*)current->n, not_node);
       // replaceNode((parse_node*)current->n, not_node, blank);
        set->push(not_node->index, (void*)not_node);
      }


   // Place the node in the Hash table
    } else {

      loc = getHashCode((parse_node*)current->n);
      current_hash = hash[loc];
      while (current_hash != NULL) {

        if (isEquivalentNode((parse_node *)current_hash->n,
                             (parse_node *)current->n)) {

          replaceNode((parse_node*)current->n, (parse_node*)current_hash->n, blank);
          set->push(((parse_node*)current_hash->n)->index, current_hash->n);
          break;

        }
        current_hash = current_hash->next;
      }



     // If the node does not match a node in the Hash table, add it
      if (((parse_node*)current->n)->index != -1) {
        current_hash = createLinkNode((parse_node*)current->n);
        current_hash->next = hash[loc];
        hash[loc] = current_hash;
      }

    }


    current = current->next;
  }



  return NOCLAUSE_OK;
}









/******************************************************************
  IsEquivalentNode
 ******************************************************************
 
  Description : Verify that 2 nodes are equivalent subformulas by
                comparing their types and children
  Parameters  : n1 - pointer to the first node to compare
                n2 - pointer to the second node to compare
  Returns     : TRUE if the nodes are equivalent
                FALSE otherwise
  Notes       : Optimize this function

 ******************************************************************/
int isEquivalentNode(parse_node *n1, parse_node *n2) {

  link_node *child;

 // Verify the nodes are the same type (i.e. AND nodes)
  if ((n1->type != n2->type) || (n1->num_children != n2->num_children))
    return FALSE;

 // Check that n1's children are the same as n2's
  child = n1->children;
  while (child != NULL) {
    if (findNode(child->n, n2->children) == FALSE)
      return FALSE;
    child = child->next;
  }

  return TRUE;
}









/******************************************************************
  FindNode
 ******************************************************************
 
  Description : Find a given node inside a linked list of nodes.
                This function is used to determine whether 2 nodes
                have the same children.
  Parameters  : source - pointer to a linked list node containing
                         the requested node
                list   - list of linked nodes that may contain
                         the nodes
  Returns     : TRUE if the node is found
                FALSE otherwise

 ******************************************************************/
int findNode(void *source, link_node *list) {

  while (list != NULL) {
    if (list->n == source) {
      return TRUE;
    }
    list = list->next;
  }

  return FALSE;
}









/******************************************************************
  ReplaceNode
 ******************************************************************
 
  Description : Merge two identical nodes
  Parameters  : replaced - the node that will be deleted
                fixed    - the new node that will represent both
                blank    - a blank unique stack for identifying the
                           parents that are the same between the
                           node to be replaced
  Modifies    : the DAG contained in root may have some nodes merged
                together (identical subtrees are joined)
  Returns     : NOCLAUSE_OK if all went well
                NOCLAUSE_ERROR otherwise

 ******************************************************************/
int replaceNode(parse_node *replaced, parse_node *fixed, UniqueInfoStack *blank) {

  link_node *current_parent, *current_child, *tmp_parent;


 // Unsure whether this condition should ever happen or not...
  if (replaced == fixed) {
    if (g_args.verbose >= 2) fprintf(stdout, "Warning: trying to replace a node with itself\n");
    return NOCLAUSE_OK;
  }


 // Since the nodes have the same children, remove all the children's
 // parent links to the replaced node
  current_child = replaced->children;
  while (current_child != NULL) {
    removeParentTwin(current_child);
    current_child = current_child->next;
  }


 // Make a list of fixed's parent nodes
  current_parent = fixed->parents;
  while (current_parent != NULL) {
    blank->push(((parse_node*)current_parent->n)->index, (void*)current_parent);
    current_parent = current_parent->next;
  }


 // Check each of replaced's nodes against this list
  current_parent = replaced->parents;
  while (current_parent != NULL) {

   // If fixed already has this parent, delete the link
    if ( blank->member(((parse_node*)current_parent->n)->index) ) {

      removeChildTwin(current_parent);

     // IFF/XOR nodes delete both links
      if ( (((parse_node*)current_parent->n)->type == IFF_TYPE) ||
           (((parse_node*)current_parent->n)->type == XOR_TYPE) ) {
        tmp_parent = (link_node*) blank->getInfo(((parse_node*)current_parent->n)->index);
        removeChildTwin(tmp_parent);
        fixed->num_parents--;
        if (tmp_parent->next != NULL) {
          tmp_parent->next->prev = tmp_parent->prev;
        }
        if (tmp_parent->prev == NULL) {
          fixed->parents = fixed->parents->next;
        } else {
          tmp_parent->prev->next = tmp_parent->next;
        }
      }
      current_parent = current_parent->next;
      

   // If fixed does not have this parent, add it to the list
    } else {
      tmp_parent = current_parent;
      current_parent = current_parent->next;
      tmp_parent->twin->n = fixed;
      tmp_parent->prev = NULL;
      fixed->parents->prev = tmp_parent;
      tmp_parent->next = fixed->parents;
      fixed->parents = tmp_parent;
      fixed->num_parents++;
    }
  }

  blank->empty();
  replaced->index = -1;
  switch (replaced->type) {
    case AND_TYPE:  g_desc.num_and_nodes--;   break;
    case OR_TYPE:   g_desc.num_or_nodes--;    break;
    case NOT_TYPE:  g_desc.num_not_nodes--;   break;
  }
  g_num_nodes--;


  return NOCLAUSE_OK;
}









/******************************************************************
  ReplaceNotNode
 ******************************************************************

  Description : This function has been added since it was determined
                that in s-exprs, it was found to be much quicker
                only for replacing NOT nodes (only important for
                large s-exprs)
  Parameters  : replaced - the node to be removed/replaced
                fixed    - the identical node that is kept
  Modifies    : Moves the parents from replaced to fixed
  Assumptions : replaced and fixed have the same children
  Returns     : NOCLAUSE_OK

 ******************************************************************/
int replaceNotNode(parse_node *replaced, parse_node *fixed) {

  link_node *current_parent;

 // Since the nodes have the same children, remove all the children's
 // parent links to the replaced node
  removeParentTwin(replaced->children);


  current_parent = replaced->parents;
  while (current_parent->next != NULL) {
    current_parent->twin->n = fixed;
    current_parent = current_parent->next;
  }
  current_parent->twin->n = fixed;
  current_parent->next = fixed->parents;
  fixed->parents->prev = current_parent;
  fixed->parents = replaced->parents;
  fixed->num_parents += replaced->num_parents;
  replaced->index = -1;

  g_desc.num_not_nodes--;
  g_num_nodes--;
  return NOCLAUSE_OK;
}
  






















/********************** Compacting Functions **********************/


/******************************************************************
  CompactRepresentation
 ******************************************************************
 
  Description : Perform a BFS of the tree converting each node to
                a compact representation (contiguous block of mem)
  Parameters  : <none>
  Modifies    : Allocates alot of memory for the problem
  Assumptions : The DAG has been parsed and built correctly
  Returns     : NOCLAUSE_OK if all went well
                NOCLAUSE_ERROR otherwise

 ******************************************************************/
int compactRepresentation() {

  int num_entries, *root, *empty;
  int *parent_watch = new int[g_desc.max_index];
  parse_node **node_list = new parse_node*[g_num_nodes];


 // Initialize this array for Breadth-First indexing
  for (int i=0; i<g_desc.max_index; i++) {
    parent_watch[i] = -1;
  }


 // Index the nodes and place them in the database
  g_desc.num_total_children = 0;
  g_desc.num_total_parents = 0;
  if (g_args.verbose >= 4) printf(" Indexing Nodes\n");
  indexNode(g_parse_root, 0, parent_watch, node_list);
  if (g_args.verbose >= 4) printf(" Done Indexing Nodes\n");



 // Allocate memory for the nodes and structures in the DAG
  num_entries =  (2 * g_num_nodes) + (2 * g_desc.num_total_children) + (2 * (g_desc.num_and_nodes + g_desc.num_or_nodes));
  g_memblock = new MemManager ( (num_entries * sizeof(int)) + 
                   (g_num_nodes * sizeof(int*)) +
                   (g_num_nodes * sizeof(byte)) +
                   (g_num_nodes * sizeof(byte)) +
                   (g_num_nodes * sizeof(int)) +
                   (g_num_nodes * sizeof(reasoninfo)) +
                   (g_num_nodes * sizeof(int) * 2) +
                   (g_num_nodes * sizeof(int)) +
                   (g_num_nodes * sizeof(Vector*)) +
                   (g_num_nodes * sizeof(Vector*)) +
                   (g_num_nodes * sizeof(Vector*)) +
                   (g_num_nodes * sizeof(Vector*)) +
                   (g_num_nodes * sizeof(Vector*)));
  assert(g_memblock);


 // Allocate memory for all the other structures
  root = (int*) g_memblock->calloc (num_entries, sizeof(int));
  g_nodes.loc = (int**) g_memblock->calloc (g_num_nodes, sizeof(int*));
  g_nodes.type = (byte*) g_memblock->calloc (g_num_nodes, sizeof(byte));
  g_nodes.value = (byte*) g_memblock->calloc (g_num_nodes, sizeof(byte));
  g_nodes.decision_level = (int*) g_memblock->calloc (g_num_nodes, sizeof(int));
  g_nodes.reason = (reasoninfo *) g_memblock->calloc (g_num_nodes, sizeof(reasoninfo));
  g_nodes.score = (int*) g_memblock->calloc (g_num_nodes * 2, sizeof(int));
  g_nodes.score_index = (int*) g_memblock->calloc (g_num_nodes, sizeof(int));
  g_nodes.true_watchlist = (Vector**) g_memblock->calloc (g_num_nodes, sizeof(Vector*));
  g_nodes.false_watchlist = (Vector**) g_memblock->calloc (g_num_nodes, sizeof(Vector*));
  g_nodes.and_watchlist = (Vector**) g_memblock->calloc (g_num_nodes, sizeof(Vector*));
  g_nodes.or_watchlist = (Vector**) g_memblock->calloc (g_num_nodes, sizeof(Vector*));
  g_nodes.dc_watchlist = (Vector**) g_memblock->calloc (g_num_nodes, sizeof(Vector*));


 // Allocate the watch vectors (nogood watches allocated in nogood.cpp)
  for (int i=0; i<g_num_nodes; i++) {
    g_nodes.true_watchlist[i] = NULL;
    g_nodes.false_watchlist[i] = NULL;
    g_nodes.and_watchlist[i] = new Vector(g_memblock, VECTOR_INITIAL_SIZE);
    g_nodes.or_watchlist[i] = new Vector(g_memblock, VECTOR_INITIAL_SIZE);
    g_nodes.dc_watchlist[i] = new Vector(g_memblock, VECTOR_INITIAL_SIZE);
  }



 // Write in all the nodes to main memory
  if (g_args.verbose >= 4) printf(" Writing Nodes\n");
  g_atom_table = new HashTable(g_memblock, PARSE_MAX_HASH_SPOTS);
  empty = root;
  for (int i=0; i<g_num_nodes; i++) {
    empty += convertNode(node_list[i], i, empty);
  }
  assert(root + num_entries == empty);
  if (g_args.verbose >= 4) printf(" Done Writing Nodes\n");




 // Set the watches for all the AND/ORs and DCs
  for (int i=0; i<g_num_nodes; i++) {

    if (g_nodes.type[i] == AND_TYPE) {
      setWatchLocation(i, 0, g_nodes.and_watchlist[ getChildNode(i,0) ]->add(i));
      setWatchLocation(i, 1, g_nodes.and_watchlist[ getChildNode(i,1) ]->add(i));
    } else if (g_nodes.type[i] == OR_TYPE) {
      setWatchLocation(i, 0, g_nodes.or_watchlist[ getChildNode(i,0) ]->add(i));
      setWatchLocation(i, 1, g_nodes.or_watchlist[ getChildNode(i,1) ]->add(i));
    }

   // All nodes but the root must have a DC watch
    if (i != 0) {
      g_nodes.dc_watchlist[ getParentNode(i,0) ]->add(i);
    }

  }


  delete[] parent_watch;
  delete[] node_list;
  return NOCLAUSE_OK;
}











/******************************************************************
  IndexNode
 ******************************************************************
 
  Description : Give an index to this node.  The index of the nodes
                must initially be -1 (as set by the createNode fcn)
  Parameters  : n            - pointer to the node to index
                index        - available index for this node
                parent_watch - extra array to make sure that we only
                               index nodes that have all their
                               parents indexed (for Breadth-First
                               ordering)
                node_list    - list of nodes that have been indexed
  Returns     : The next valid index (same or incremented by 1)

 ******************************************************************/
int indexNode(parse_node *n, int index, int *parent_watch, parse_node **node_list) {

  link_node *current;
  parse_node *child;
  int num_children = 0, num_parents = 0;


 // Ignore nodes that have already been assigned
  if (n->index < 0)
    return index;


 // Set the index for this node (mark that it has been written)
  node_list[index] = n;
  n->index = -(++index);


 // Set the index of all of this node's children
  current = n->children;
  while (current != NULL) {
    child = (parse_node*) current->n;
    if (parent_watch[child->index] == -1)
      parent_watch[child->index] = child->num_parents;
    parent_watch[child->index]--;
    num_children++;
    current = current->next;
  }


 // Index any children that have all their parents indexed
  current = n->children;
  while (current != NULL) {
    child = (parse_node*) current->n;
    if (parent_watch[child->index] == 0)
      index = indexNode(child, index, parent_watch, node_list);
    current = current->next;
  }


 // Count the number of parents (double check debugging)
  current = n->parents;
  while (current != NULL) {
    num_parents++;
    current = current->next;
  }


 // Add the number of children to the global total
  assert(num_children == n->num_children);
  assert(num_parents == n->num_parents);
  g_desc.num_total_children += n->num_children;
  g_desc.num_total_parents += n->num_parents;
  if (n->num_children > g_desc.max_children)
    g_desc.max_children = n->num_children;
  if (n->num_parents > g_desc.max_parents)
    g_desc.max_parents = n->num_parents;

 // Update some more counters
  if (n->type == AND_TYPE) {
    g_desc.num_and_children += n->num_children;
    if (n->num_children > g_desc.max_and_children)
      g_desc.max_and_children = n->num_children;
  } else if (n->type == OR_TYPE) {
    g_desc.num_or_children += n->num_children;
    if (n->num_children > g_desc.max_or_children)
      g_desc.max_or_children = n->num_children;
  }


  return index;
}






/******************************************************************
  ConvertNode
 ******************************************************************
 
  Description : Add a node to the next available location in the
                contiguous block of memory as identified by
		g_empty_loc.
  Parameters  : n     - pointer to the node to convert (non-null!)
                index - the current index (debugging)
                loc   - available location for this node
  Modifies    : For efficiency considerations, this function destroys
                the tree as it converts it.
  Assumes     : No nodes are of type UNDEFINED_TYPE
  Returns     : The next valid location (incremented by the size
                of this node)

 ******************************************************************/
int convertNode(parse_node *n, int index, int *loc) {

  int pos;

 // Debugging
  assert(index == (-1) - n->index);


 // Set the basic information
  loc[0] = n->num_children;
  loc[1] = n->num_parents;
  g_nodes.type[index] = (byte) n->type;
  if (n->type == ATOM_TYPE) {
    g_atom_table->add(n->identifier, index);
  }

  g_nodes.loc[index] = loc;
  if ((n->type == IFF_TYPE) && (n->num_children == 0))
    g_nodes.value[index] = TRUE;
  else if ((n->type == XOR_TYPE) && (n->num_children == 0))
    g_nodes.value[index] = FALSE;
  else
    g_nodes.value[index] = UNKNOWN;


 // Write the indices of all the children
  pos = 2;
  while (n->children != NULL) {
    loc[pos++] = (-1) - (((parse_node*)(n->children->n))->index);
    n->children = n->children->next;
  }


 // Write the indices of all the parents
  while (n->parents != NULL) {
    loc[pos++] = (-1) - (((parse_node*)(n->parents->n))->index);
    n->parents = n->parents->next;
  }
  assert(pos-2-n->num_children == (int)n->num_parents);


 // Leave some room for AND and OR watches
  if ((n->type == AND_TYPE) || (n->type == OR_TYPE)) {
    assert(n->num_children >= 2);
    pos += 2;
  }


  return pos;
}











/*********** Generic Node Creation/Deletion Functions *************/


/******************************************************************
  CreateAtom
 ******************************************************************
 
  Description : Allocate a new child atom or finds it if it already
                exists.
  Parameters  : identifier - a pointer to the identifier string
  Modifies    : Allocates memory for the new atom node
  Returns     : A pointer to the new atom
  Note        : All atoms should be found/created using this
                function.

 ******************************************************************/
parse_node *createAtom(char *id) {

  int loc;
  parse_node *newatom = NULL;
  link_node *list, *newchildlink;


 // Look for the node in the hash table
  assert(id);
  loc = getVariableLocation(id);
 /* printf("Id is %s placed at %d out of %d\n", id, loc, PARSE_MAX_HASH_SPOTS); */
  list = g_hashed_vars[loc];
  while (list != NULL) {
    if (strcasecmp(id, ((parse_node*)(list->n))->identifier) == 0) {
      newatom = (parse_node*)(list->n);
      break;
    }
    list = list->next;
  }


 // Otherwise, create a new node
  if (newatom == NULL) {

    newatom = createNode();
    newatom->type = ATOM_TYPE;
    newatom->identifier = (char*) g_tmp_memblock->calloc(strlen(id)+1, sizeof(char));
    strcpy(newatom->identifier, id);
    g_desc.num_atoms++;

  // Add the new atom to the variable list
    newchildlink = createLinkNode(newatom);
    newchildlink->next = g_hashed_vars[loc];
    if (g_hashed_vars[loc] != NULL)
      g_hashed_vars[loc]->prev = newchildlink;
    g_hashed_vars[loc] = newchildlink;
  }


  return newatom;
}








/******************************************************************
  CreateNode
 ******************************************************************
 
  Description : Allocate a new node in memory
  Parameters  : <none>
  Modifies    : Allocates memory for the new node
  Returns     : A pointer to the new node

 ******************************************************************/
parse_node *createNode() {

  parse_node *newnode = (parse_node*) g_tmp_memblock->calloc(1, sizeof(parse_node));
  assert(newnode);

 // Initialize the node values to their defaults
  newnode->type = UNDEFINED_TYPE;
  newnode->index = g_desc.max_index++;
  newnode->num_children = 0;
  newnode->num_parents = 0;
  newnode->identifier = NULL;
  newnode->children = NULL;
  newnode->parents = NULL;

 // Increase the global node count
  g_num_nodes++;
  
  return newnode;
}








/******************************************************************
  CreateLinkNode
 ******************************************************************
 
  Description : Allocate a new linked list node in memory
  Parameters  : n   - a pointer to the node represented by this
                      linked list entry
  Modifies    : Allocates memory for the new linked node
  Returns     : A pointer to the new linked node

 ******************************************************************/
link_node *createLinkNode(parse_node *n) {

  link_node *newlink = (link_node*) g_tmp_memblock->calloc (1, sizeof(link_node));
  assert(newlink);
  
 // Initialize the node values to their defaults
  newlink->n = (void*)n;
  newlink->twin = NULL;
  newlink->next = NULL;
  newlink->prev = NULL;
  
  return newlink;
}







/******************************************************************
  RemoveParentTwin
 ******************************************************************
 
  Description : Remove the paired/twin link node
  Parameters  : n - link node whose twin should be removed
  Returns     : <none>

 ******************************************************************/
void removeParentTwin(link_node *n) {

  link_node *twin = n->twin;

  ((parse_node*)n->n)->num_parents--;
  if (twin->prev == NULL) {
      ((parse_node*)n->n)->parents = twin->next;
  } else {
      twin->prev->next = twin->next;
  }
  if (twin->next != NULL)
    twin->next->prev = twin->prev;
}




/******************************************************************
  RemoveChildTwin
 ******************************************************************
 
  Description : Remove the paired/twin link node
  Parameters  : n - link node whose twin should be removed
  Returns     : <none>

 ******************************************************************/
void removeChildTwin(link_node *n) {

  link_node *twin = n->twin;

  ((parse_node*)n->n)->num_children--;
  if (twin->prev == NULL) {
      ((parse_node*)n->n)->children = twin->next;
  } else {
      twin->prev->next = twin->next;
  }
  if (twin->next != NULL)
    twin->next->prev = twin->prev;
}




































/******************** Other Parsing Functions *********************/
// The S-Expr function is incomplete and should be fixed



/******************************************************************
  BuildParseTreeFromCNF
 ******************************************************************
 
  Description : Parse the CNF formula contained in the file and
                create an internal parse tree.
  Parameters  : filename - null-terminated string containing the
                           location of the file.
  Returns     : NOCLAUSE_OK if all went well
                NOCLAUSE_FILE_NOT_FOUND if the file cannot be opened

 ******************************************************************/
int buildParseTreeFromCNF(char *filename) {


  FILE *infile;
  char inputline[PARSE_MAX_STRLEN];
  char *token, *delimiters = " \n\t";
  int num_literals, num_clauses, literalvalue;
  parse_node *clause, *literal, *negation;



 // Ouput a status message
  if (g_args.verbose >= 3) {
    fprintf(stdout, "Parsing CNF problem contained in file \"%s\"\n",
            filename);
  }


 // Make sure the DAG is empty and open the input file
  assert(!g_parse_root);
  infile = openInputFile(filename);
  if (infile == NULL) {
    return NOCLAUSE_FILE_NOT_FOUND;
  }



 // Find the number of variables and clauses
  num_literals = 0;
  num_clauses = 0;

  while (fgets(inputline, PARSE_MAX_STRLEN, infile) != NULL) {
    if ((inputline[0] == 'p') || (inputline[0] == 'P')) {
      token = strtok(inputline, delimiters);
      token = strtok(NULL, delimiters);
      if (strcasecmp(token, "cnf") == 0) {
        num_literals = strtol(strtok(NULL, delimiters), NULL, 10);
        num_clauses = strtol(strtok(NULL,delimiters), NULL, 10); 
        if (g_args.verbose >= 2) printf("Found %d literals and %d clauses\n",
               num_literals, num_clauses);
        break;
      }
    }
  }


 // Check that we found the number of literals and clauses
  if ((num_literals == 0) || (num_clauses == 0)) {
    fprintf(stderr, "Error: could not find number of clauses/variables.\n");
    return NOCLAUSE_PARSE_ERROR;
  }



 // Build the top level AND clause
  g_parse_root = createNode();
  g_parse_root->type = AND_TYPE;
  g_desc.num_and_nodes++;
  clause = createNode();
  clause->type = OR_TYPE;



 // Create new nodes for each clause
  while ((g_parse_root->num_children < num_clauses) && 
         (fgets(inputline, PARSE_MAX_STRLEN, infile) != NULL)) {

    token = strtok(inputline, delimiters);
    while (token != NULL) {

     // Ignore comments
      if ((token[0] == 'c') || (token[0] == 'C') ||
          (token[0] == 'p') || (token[0] == 'P') ||
          (token[0] == '%')) {
        break;


      } else {

        literalvalue = strtol(token, NULL, 10);
        if ((literalvalue == 0) && (clause->num_children > 0)) {

         // Add the clause to the tree
          if (clause->num_children == 1) {
            literal = (parse_node*) (clause->children->n);
            int num_parents = literal->num_parents;
            removeParentTwin(clause->children);
            assert(num_parents-1 == literal->num_parents);
            addChild(g_parse_root, literal);
            clause->num_children = 0;
            clause->children = NULL;
          } else {
            addChild(g_parse_root, clause);
            g_desc.num_or_nodes++;
            clause = createNode();
            clause->type = OR_TYPE;
          }


       // Add a new literal to the temporary clause
        } else {
          if (literalvalue < 0) {
            literal = createAtom(&token[1]);
            negation = createNode();
            negation->type = NOT_TYPE;
            g_desc.num_not_nodes++;
            addChild(negation, literal);
            addChild(clause, negation);
          } else {
            literal = createAtom(token);
            addChild(clause, literal);
          }
        }
      }

      token = strtok(NULL, delimiters);
    } // while (token != NULL)
  } // while (fgets(...) != NULL)



 // Output a warning if we reached EOF before finding all expected clauses
  if (g_parse_root->num_children < num_clauses) {
    fprintf(stderr, "Warning: Expected %d clauses but found %d\n",
            num_clauses, g_parse_root->num_children);
  }


 // Close the file
  fclose(infile);
  g_num_nodes--;
  return NOCLAUSE_OK;
}












/******************************************************************
  BuildParseTreeFromSexpr
 ******************************************************************
 
  Description : Parse the S-expression contained in the file and
                create an internal parse tree.
  Parameters  : filename - null-terminated string containing the
                           location of the file.
  Returns     : NOCLAUSE_OK if all went well
                NOCLAUSE_PARSE_ERROR if the file does not contain a
                valid S-expression
                NOCLAUSE_ERROR otherwise

 ******************************************************************/
int buildParseTreeFromSexpr(char *filename) {

  FILE *infile;

 /* Ouput a status message */
  if (g_args.verbose >= 3) {
    fprintf(stdout, "Parsing S-expression contained in file \"%s\"\n",
            filename);
  }


 /* Make sure the DAG is empty and open the input file */  
  assert(!g_parse_root);
  infile = openInputFile(filename);
  if (infile == NULL) {
    return NOCLAUSE_FILE_NOT_FOUND;
  }


 /* Build the tree recursively */
  g_parse_root = buildNodeFromSexpr(infile);
  if (g_parse_root == NULL) {
    return NOCLAUSE_PARSE_ERROR;
  }

  fclose(infile);
  return NOCLAUSE_OK;
}








parse_node *buildNodeFromSexpr(FILE *infile) {

  int retval;
  char token[PARSE_MAX_TOKLEN];
  parse_node *n, *child;


 /* Get the node */
  retval = getToken(infile, token);
  if (retval == PARSE_STRING_TOKEN) {
    return createAtom(token);
  } else if (retval != PARSE_OPEN_BRACKET_TOKEN) {
    return NULL;
  }
  

 /* Otherwise, build a new node if we find a '(' */
  retval = getToken(infile, token);
  if (retval != PARSE_STRING_TOKEN) {
    return NULL;
  }


 /* Create the new node and look for the type */
  n = createNode();
  if        (strcasecmp("AND", token) == 0) {  n->type = AND_TYPE;
  } else if (strcasecmp("OR" , token) == 0) {  n->type = OR_TYPE;
  } else if (strcasecmp("NOT", token) == 0) {  n->type = NOT_TYPE;
  } else if (strcasecmp("XOR", token) == 0) {  n->type = XOR_TYPE;
  } else if (strcasecmp("IFF", token) == 0) {  n->type = IFF_TYPE;
  } else { return NULL; }


 /* Add all the children to the node */
  while ((child = buildNodeFromSexpr(infile)) != NULL) {
    n = combineChild(n, child);
  }


 /* Make sure only NOTs have 1 child */
  assert(n->num_children > 0);
  assert((n->type != NOT_TYPE) || (n->num_children == 1));
  if ((n->type != NOT_TYPE) && (n->num_children == 1)) {
    ((parse_node*)n->children->n)->parents = NULL;
    ((parse_node*)n->children->n)->num_parents = 0;
    g_num_nodes--;
    return ((parse_node*)n->children->n);
  }


  return n;
}






parse_node *combineChild(parse_node *n, parse_node *child) {


 /* Change NOT-type-NOT to the twin type */
  if (n->type == NOT_TYPE) {

    switch (child->type) {

    case NOT_TYPE:
      ((parse_node*)child->children->n)->parents = NULL;
      ((parse_node*)child->children->n)->num_parents = 0;
      g_num_nodes -= 2;
      return ((parse_node*)child->children->n);

    case AND_TYPE:
      if (countNegativeChildren(child) > child->num_children/2) {
        child->type = OR_TYPE;
        swapNegativeChildren(child);
        g_num_nodes--;
        return child;
      }
      return addChild(n, child);

    case IFF_TYPE:
      if (countNegativeChildren(child) > child->num_children/2) {
        child->type = XOR_TYPE;
        swapNegativeChildren(child);
        g_num_nodes--;
        return child;
      }
      return addChild(n, child);

    case OR_TYPE:
      if (countNegativeChildren(child) > child->num_children/2) {
        child->type = AND_TYPE;
        swapNegativeChildren(child);
        g_num_nodes--;
        return child;
      }
      return addChild(n, child);

    case XOR_TYPE:
      if (countNegativeChildren(child) > child->num_children/2) {
        child->type = IFF_TYPE;
        swapNegativeChildren(child);
        g_num_nodes--;
        return child;
      }
      return addChild(n, child);

    default:
      return addChild(n, child);
    }
  }


 /* Combine nested expressions */      
  if ( (n->type == child->type) &&
      ((n->type == OR_TYPE) || (n->type == AND_TYPE) ||
       (n->type == XOR_TYPE) || (n->type == IFF_TYPE)) ){
    g_num_nodes--;
    return moveChildren(n, child->children);
  }

 /* Otherwise just add the child to the node */
  return addChild(n, child);
}








/******************************************************************

 ******************************************************************

  Description :
  Parameters  : <none>
  Modifies    : <none>
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
int getToken(FILE *infile, char *s) {

  char input;
  int length = 0;
  static int stored = PARSE_NO_TOKEN;


 /* Look for a previously stored token */
  if (stored == PARSE_OPEN_BRACKET_TOKEN) {
    stored = PARSE_NO_TOKEN;
    return PARSE_OPEN_BRACKET_TOKEN;

  } else if (stored == PARSE_CLOSE_BRACKET_TOKEN) {
    stored = PARSE_NO_TOKEN;
    return PARSE_CLOSE_BRACKET_TOKEN;
  }


 /* Read in the next token */
  while ((input = fgetc(infile)) != EOF) {

   /* Check that the token is not too long */
    if (length >= PARSE_MAX_TOKLEN-1) {
      return NOCLAUSE_PARSE_ERROR;


   /* If we already have a string, look for termination */
    } else if (length > 0) {
    
      switch (input) {
      case '\n':
      case '\t':
      case '\r':
      case '\f':
      case '\v':
      case ' ': s[length++] = '\0';
                return PARSE_STRING_TOKEN;
      case ')': s[length++] = '\0';  
                stored = PARSE_CLOSE_BRACKET_TOKEN;
                return PARSE_STRING_TOKEN;
      case '(': s[length++] = '\0';
                stored = PARSE_OPEN_BRACKET_TOKEN;
                return PARSE_STRING_TOKEN;
      default:  s[length++] = input;
      }


   /* Look for a single special character */
    } else {

      switch (input) {
      case '\n':
      case '\t':
      case '\r':
      case '\f':
      case '\v':
      case ' ': continue;
      case ')': return PARSE_CLOSE_BRACKET_TOKEN;
      case '(': return PARSE_OPEN_BRACKET_TOKEN;
      default:  s[length++] = input;
      }

    }
   
  } /* End While */


  return NOCLAUSE_EOF_ERROR;
}






int countNegativeChildren(parse_node *n) {

  int count = 0;
  link_node *current = n->children;

  while (current != NULL) {
    if (((parse_node*)current->n)->type == NOT_TYPE)
      count++;
    current = current->next;
  }

  return count;
}





void swapNegativeChildren(parse_node *n) {

  int i, pos_count = 0, neg_count = 0;
  link_node *pos_children[n->num_children];
  link_node *neg_children[n->num_children];
  link_node *current = n->children;
  link_node *twin;
  void *swap;

 /* Find all the positive and negative children */
  while (current != NULL) {
    if (((parse_node*)current->n)->type == NOT_TYPE) {
      neg_children[neg_count++] = current;
    } else {
      pos_children[pos_count++] = current;
    }
    current = current->next;  
  }


 /* For each positive child, move one of the negatives */
  for (i=0; i<pos_count; i++) {
    swap = pos_children[i]->n;
    twin = pos_children[i]->twin;

    current = ((parse_node*)neg_children[i]->n)->children;
    pos_children[i]->n = current->n;
    pos_children[i]->twin = current->twin;
    pos_children[i]->twin->twin = pos_children[i];
    pos_children[i]->twin->n = n;

    current->n = swap;
    current->twin = twin;
    current->twin->twin = current;
    current->twin->n = neg_children[i]->n;
  }
  

 /* Remove all the excess NOTs on the remaining negative children */
  for (i=pos_count; i<neg_count; i++) {
    current =((parse_node*)neg_children[i]->n)->children;
    neg_children[i]->n = current->n;
    neg_children[i]->twin = current->twin;
    neg_children[i]->twin->twin = neg_children[i];
    neg_children[i]->twin->n = n;
    g_num_nodes--;
  }

}





parse_node *moveChildren(parse_node *n, link_node *child_list) {

  int count = 1;
  link_node *current = child_list;


 /* Do nothing if we have an empty list */
  if (child_list == NULL)
    return n;


 /* Transfer the parent pointers in the child list */
  while (current->next != NULL) {
    current->twin->n = (void*)n;
    count++;
    current = current->next;  
  }
  current->twin->n = (void*)n;


 /* Add the list to n */
  if (n->children != NULL) {
    n->children->prev = current;
    current->next = n->children;
  }
  n->children = child_list;
  n->num_children += count;

  return n;
}












/********************** Hash Table Functions **********************/
// TO BE REPLACED WITH THE HASH TABLE OBJECT...


/******************************************************************
  CreateVariableHashTable
 ******************************************************************
 
  Description : Add the variable in the leaf to the hashing table.
                This function purposely adds multiple duplicates
	        since they will later be used to find identical 
		subformulas.
  Parameters  : <none>
  Returns     : <none>

 ******************************************************************/
void createVariableHashTable() {

  int i;

  /* Initialize all the entries to NULL */
  for (i=0; i<PARSE_MAX_HASH_SPOTS; i++) {
    g_hashed_vars[i] = NULL;
    g_hashed_iscas_nodes[i] = NULL;
  }

}







/******************************************************************
  DeleteVariableHashTable
 ******************************************************************
 
  Description : Delete all the DPLL nodes in the hash table
  Parameters  : <none>
  Returns     : <none>

 ******************************************************************/
void deleteVariableHashTable() {

  int i, non_empty = 0;

  
 // Remove all the linked list entries
  for (i=0; i<PARSE_MAX_HASH_SPOTS; i++) {
    if (g_hashed_vars[i] != NULL) {
      g_hashed_vars[i] = NULL;
      non_empty++;
    }
  }


 // Output the percentage complete of the hash table
  if (g_args.verbose >= 2) {
    fprintf(stdout, "Variable to Hash entry ratio is %3.2f%%\n",
            (((double)g_desc.num_atoms) / PARSE_MAX_HASH_SPOTS) * 100);
    fprintf(stdout, "Hash table is %3.2f%% full\n",
            (((double)non_empty) / PARSE_MAX_HASH_SPOTS) * 100);
  }

}






/******************************************************************
  GetVariableLocation
 ******************************************************************
 
  Description : Finds a non-negative integer representing the
                variable for this integer.  This function guarantees
		to always return the same identifier for pairwise
		identical strings.  Because this function uses
		hashing, the integers are not guaranteed to be
		unique.
  Parameters  : s - the string representing this identifier as read
                    from the file
  Returns     : An integer representing this variable
  Notes       : Fix the hashing function!

 ******************************************************************/
unsigned int getVariableLocation(char *s) {

  unsigned int loc;
  int i, max;
  assert(s);

 // Fix the HASHING function!
  loc = 1;
  max = strlen(s);
  for (i=0; i<max; i++) {
    loc *= ((int) (s[i])) * (i+1);
  }
  loc %= PARSE_MAX_HASH_SPOTS;


 // Output the location of the variable for debugging
  if (g_args.verbose >= 5) {
    fprintf(stdout, " Found variable location %d for \"%s\" of length %d\n",
            loc, s, max);
  }
  return loc;
}





