/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  June 3rd 2003 (April 2004)

  *****************************************************************

  Parser.h

  Contains the declarations of the functions required to parse an
  S-expression written to a file into an internal DAG representation.

  Prime numbers chosen from
    www.utm.edu/research/primes/lists/small/1000.txt

  *****************************************************************/

#ifndef __PARSER_H__
#define __PARSER_H__
  
#include "noclause.h"


/************************* Constants ******************************/

#define PARSE_DEFAULT_BLOCKSIZE  16777211
#define PARSE_MAX_STRLEN             1024
#define PARSE_MAX_TOKLEN              256
#define PARSE_OPEN_BRACKET_TOKEN        1
#define PARSE_CLOSE_BRACKET_TOKEN       2
#define PARSE_STRING_TOKEN              3
#define PARSE_NO_TOKEN                  0

#define PARSE_MAX_HASH_SPOTS         7919
#define PARSE_MERGING_HASH_CODE      4999

#define PARSE_TOKEN_DEFAULT             0
#define PARSE_TOKEN_EOLN                1
#define PARSE_TOKEN_INPUT               2
#define PARSE_TOKEN_OUTPUT              3



/****************** Structure Declarations ************************/


/* Linked-List node definition
    - pointer to a node
    - pointer to the corresponding child/parent link
    - pointer to the next linked list entry
    - pointer to the previous linked list entry
 */
typedef struct link_node {

  void *n;
  struct link_node *twin;
  struct link_node *next;
  struct link_node *prev;
  
} link_node;




/* Node Structure used in parsing
    - unique index identifier
    - number of children (for quick reference)
    - number of parents (for quick reference)
    - node type (i.e. AND, NOT, OR, etc...)
    - string identifier for atom nodes
    - linked list of children
    - linked list of parents
 */
typedef struct parse_node {

  int index;
  int num_children;
  int num_parents;
  unsigned short type;
  char *identifier;
  link_node *children;
  link_node *parents;
  
} parse_node;







/* This is used to read an ISCAS file for parsing and verification
   purposes.  The object takes an input file and can be used to
   identify token types (INPUT/OUTPUT signals) and automatically
   separates the various tokens (by identifying spaces, brackets,
   commas, etc..).  An advantage of this object is that the entire
   line may be very long (hundreds of thousands of characters) and
   the tokenizer will only read parts of the line at a time, hiding
   all the details from the main program such that the program only
   has to call for another token, without worrying about reading a
   new line and making sure a token name does not intersect both the
   lines.
*/
class ISCASTokenizer {

protected:

  FILE *infile;
  int pos, len;
  char inputline[PARSE_MAX_STRLEN];
  char token[PARSE_MAX_TOKLEN];
  char output[PARSE_MAX_TOKLEN];


 // Read the next line, up to EOLN or MAX_STRLEN
  void inline readLine() {
    char *tmp;
    do {
      tmp = fgets(inputline, PARSE_MAX_STRLEN, infile);
    } while ((tmp != NULL) && (inputline[0] == '#'));
    pos = 0;
    if (tmp == NULL)    len = 0;
    else                len = strlen(inputline);
  }


 // Skip any white space in the current input line
  void inline skipWhiteSpace() {
    while ((pos < len) && ((inputline[pos] == ' ') ||
           (inputline[pos] == '\t') || (inputline[pos] == '\n') ||
           (inputline[pos] == ',')  || (inputline[pos] == '=')  ||
           (inputline[pos] == '('))) {
      pos++;
    }
  }


 // Return NULL if eoln (and read a new line) or get the next token
 // in the line
  char inline *readToken() {

    int count = 0;

   // Find the start of the next token, if EOLN, get the new line
    skipWhiteSpace();
    if (pos >= len) {
      readLine();
      if (len == 0) {
        return NULL;
      }
      skipWhiteSpace();
    }

   // If there is another token, check if the token overlaps two
   // lines
    do {
      token[count++] = inputline[pos++];
      if (pos >= len)
        readLine();
    } while ((pos < len) && (inputline[pos] != ' ') &&
             (inputline[pos] != '\t') && (inputline[pos] != '\n') &&
             (inputline[pos] != ',')  && (inputline[pos] != '=')  &&
             (inputline[pos] != '(')  && (inputline[pos] != ')'));
    token[count] = '\0';

   // If the token is just a closing bracket, then we know we are at
   // the end of the current gate definition
    if (token[0] == ')') {
      return NULL;
    }

    return token;
  }



public:

 // Constructor opens a file
  ISCASTokenizer(char *filename) {
    infile = fopen(filename, "r");
    output[0] = '\0';
    readLine();
  }


 // Destructor
  ~ISCASTokenizer() {
    fclose(infile);
  }


 // Get the next token, write it in 's' and return a specific
 // token constant to indicate the type of signal (INPUT/OUTPUT)
 // or if we are at the end of a gate definition (EOLN)
  int inline nextToken(char **s) {
    if (infile == NULL)
      return NOCLAUSE_FILE_NOT_FOUND;

    if (readToken() == NULL) {
      return PARSE_TOKEN_EOLN;
    }

    if (strcasecmp("OUTPUT", token) == 0) {
      char *rv;

      sprintf(output, "%s", readToken());
      rv = readToken();
      assert(rv == NULL);
      rv = readToken();
      assert(rv != NULL);
    }

    if (strcasecmp("INPUT", token) == 0) {
      readToken();
      *s = token;
      return PARSE_TOKEN_INPUT;
    }

    if (strcasecmp(output, token) == 0) {
      return PARSE_TOKEN_OUTPUT;
    }

    *s = token;
    return PARSE_TOKEN_DEFAULT;
  }

};







/* Hash entry for various uses, including keeping a table
   of all the atoms for future verification
    - a string, usually the name of the atom in the input file
      the hash table uses this string to determine the location
    - additional info such as the node's index in the DAG
    - pointer to the next entry that has the same hash code
 */
typedef struct hash_entry {

  char *s;
  int info;
  struct hash_entry *next;

} hash_entry;




/* Hash table object that can be used to add or find entries that
   are referenced by a string (see hash entry for further detail on
   how the hash table is composed).  This table uses a memory
   manager so that the table is part of the allocation and will be
   efficiently deallocated by the manager.
 */
class HashTable {

protected:

  MemManager *mng;
  int size;
  hash_entry **table;


 // Hash Position function
  int inline getPos(char *s) {
    unsigned int loc = 1;
    for (int i=0; i<(int)strlen(s); i++) {
      loc *= ((int) (s[i])) * (i+1);
    }
  // printf("%s has value %d\n", s, (int) (loc % size));
    return (int) (loc % size);
  }



public:

 // Constructor, allocate memory for the table from the manager
  HashTable(MemManager *m, int max) {
    mng = m;
    table = (hash_entry**) mng->calloc (max, sizeof(hash_entry*));
    size = max;
    for (int i=0; i<size; i++)
      table[i] = NULL;
  }



 // Add an entry to the table with optional additional info
  void inline add(char *s, int info) {
    hash_entry *he = (hash_entry*) mng->calloc (1, sizeof(hash_entry));
    char *id = (char*) mng->calloc (strlen(s)+1, sizeof(char));
    sprintf(id, "%s", s);
    he->s = id;
    he->info = info;
    int pos = getPos(s);
    he->next = table[pos];
    table[pos] = he;
  }


 // Get the info corresponding to a string
  int inline getInfo(char *s) {
    int pos = getPos(s);
    hash_entry *he = table[pos];
    while (he != NULL) {
      if (strcasecmp(s, he->s) == 0)
        return he->info;
      he = he->next;
    }
    return -1;
  }

};




/********************** Global Variables **************************/

 // The atom table is kept for verification and must be deleted
 // by the main program... (should probably fix this)
extern HashTable *g_atom_table;




/************ Externally Available Function Declarations **********/

extern int Parse_ISCAS(char *filename);
extern int Parse_VerifyISCAS(char *filename);

 // Untested since alot has changed
extern int Parse_Sexpr(char *filename);
extern int Parse_CNF(char *filename);

#endif
