/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  October 18th 2003 (April 2004)

  *****************************************************************

   Propagate.cpp

   Contains the declaration of the functions required to propagate
   a set of node assignments.  Assignments are propagated because of
   nogoods, the children or the parents.  Furthermore, AND and OR
   nodes may force some children to be Don't Cares.  DC are then
   propagated (if g_args.dcpropagation == TRUE).

   Some functions in this file are inlined for efficiency, which
   forces them to be located BEFORE the functions that use them.

  *****************************************************************/

#include "propagate.h"
#include "solver.h"
#include "undo.h"
#include "nogood.h"
#include "stats.h"





/********************** Global Variables **************************/

 // Stack of assignments that are waiting to be propagated
int *g_propagation_stack = NULL;
int g_stack_top = 0;


 // Stack of parents that force their children to be DCs, the
 // UniqueStack object prevents duplicate nodes from being placed
 // on the stack.
UniqueStack *g_dcstack = NULL;






/******************* Local Function Prototypes ********************/

 // Propagate NoGoods watched by the node
int propagateNoGoods(int node);
int moveNoGoodWatch(int index, int *nogood);

 // Propagate the assignment to the children
int propagateNode(int node);
int propagateAndNode(int node);
int propagateOrNode(int node);
int propagateNotNode(int node);


 // Propagate DCs
int  propagateDCs();
void moveDCWatch(int node);


 // These functions need to be modified... (not a priority since
 //   ISCAS does not contain IFF and XOR nodes)
int propagateIffNode(int node);
int propagateXorNode(int node);
int propagateIffParent(int parent, int child);
int propagateXorParent(int parent, int child);









/***************** Externally Available Functions *****************/



/******************************************************************
  Propagate_Initialize
 ******************************************************************
 
  Description : Initialize the structures that will be used for
                propagation (namely the stacks)
  Parameters  : <none>
  Modifies    : allocates memory for the two stacks
  Assumptions : the stacks are not already allocated (in which case
                we lose the memory)
  Returns     : <none>

 ******************************************************************/
void Propagate_Initialize() {

  g_propagation_stack = (int*) malloc (g_num_nodes * sizeof(int));
  assert(g_propagation_stack);

  if (g_args.dcpropagation == TRUE) {
    g_dcstack = new UniqueStack(g_num_nodes);
    assert(g_dcstack);
  }

  Propagate_Empty();
}









/******************************************************************
  Propagate_Free
 ******************************************************************
 
  Description : Free the memory allocated to the structures
  Parameters  : <none>
  Modifies    : deallocates memory
  Assumptions : the stacks are already allocated
  Returns     : <none>

 ******************************************************************/
void Propagate_Free() {

  free(g_propagation_stack);
  if (g_args.dcpropagation == TRUE) {
    delete g_dcstack;
  }

}










/******************************************************************
  Propagate_Propagate
 ******************************************************************
 
  Description : Propagates all nodes on the propagation stack
  Parameters  : <none>
  Modifies    : removes all nodes on the stack and may make some
                assignments using Solver_Assign
  Assumptions : the structures are allocated
  Returns     : SOLVER_OK or SOLVER_CONFLICT if Solver_Assign
                returns a conflict

 ******************************************************************/
int Propagate_Propagate() {

  int node, retval;


  while (g_stack_top > 0) {

   // Get the top node on the stack
    node = g_propagation_stack[--g_stack_top];


   // Propagate any assignments inferred from NoGoods
    retval = propagateNoGoods(node);
    if (retval != SOLVER_OK) {
      Propagate_Empty();
      return retval;
    }


  // Propagate the value to the children and parents using standard
  // Boolean propagation rules
    retval = propagateNode(node);
    if (retval != SOLVER_OK) {
      Propagate_Empty();
      return retval;
    }

  }


 // Propagate the Don't Cares
  if (g_args.dcpropagation == TRUE) propagateDCs();


  return SOLVER_OK;
}
















/************************ Local Functions *************************/



/******************************************************************
  SwapWatches
 ******************************************************************
 
  Description : The watch children are the two first nodes in the
                children list.  Move the watches means that the
                children in the list must be swapped to the front.
  Parameters  : parent        - the node that is being watched
                old_watch_loc - location of the old watch child
                                (either 0 or 1)
                new_watch_loc - current location of the new watch
  Modifies    : Moves the children in the child list of 'parent',
                adds watches to the
  Assumptions : The current watches are the first two children in
                the list and the indices of where this parent is
                located in the children's watch lists are available
                (at the end of the child list)
  Returns     : <none>

 ******************************************************************/
void inline swapWatches(int parent, int old_watch_loc, int new_watch_loc) {

 // Verify some assumptions (debugging)
  assert((old_watch_loc == 0) || (old_watch_loc == 1));
  assert(new_watch_loc < getNumChildren(parent));

 // Get the location of the watch in the watch vector
  int loc = getWatchLocation(parent, old_watch_loc);
  int swapped_parent;


 // Remove the watch for the current child
  if (g_nodes.type[parent] == AND_TYPE) {
    g_nodes.and_watchlist[ getChildNode(parent, old_watch_loc) ]->remove(loc, &swapped_parent);
  } else {
    assert(g_nodes.type[parent] == OR_TYPE);
    g_nodes.or_watchlist[ getChildNode(parent, old_watch_loc) ]->remove(loc, &swapped_parent);
  }


 // Fix the watches (since vectors swap the last child during removal)
  if (swapped_parent != -1) {

    if (getChildNode(swapped_parent, 0) == getChildNode(parent, old_watch_loc)) {
      setWatchLocation(swapped_parent, 0, loc);
    } else {
      assert(getChildNode(swapped_parent, 1) == getChildNode(parent, old_watch_loc));
      setWatchLocation(swapped_parent, 1, loc);
    }

  }


 // Swap the children in the child list
  swapChildren(parent, old_watch_loc, new_watch_loc);


 // Add watches for the new watch child
  if (g_nodes.type[parent] == AND_TYPE) {
    setWatchLocation(parent, old_watch_loc, g_nodes.and_watchlist[ getChildNode(parent, old_watch_loc) ]->add(parent));
  } else {
    assert(g_nodes.type[parent] == OR_TYPE);
    setWatchLocation(parent, old_watch_loc, g_nodes.or_watchlist[ getChildNode(parent, old_watch_loc) ]->add(parent));
  }


}







/******************************************************************
  MoveChildrenWatches
 ******************************************************************
 
  Description : If the watches are assigned the correct value (TRUE
                for an AND and FALSE for an OR), then find new
                watches.
  Parameters  : parent - the parent that is being watched
                value  - the value of the watches that should be
                         moved
  Modifies    : May move some watches or assign the final child
  Assumptions : All structures are initialized with the watch
                children being the two first nodes in 'parent's
                list
  Returns     : SOLVER_OK or SOLVER_CONFLICT if the implication
                forces a conflict

 ******************************************************************/
int inline moveChildrenWatches(int parent, byte value) {

 // Get the watch children
  int first_watch = getChildNode(parent, 0);
  int second_watch = getChildNode(parent, 1);
  int i = 2, child, num_children = getNumChildren(parent);


 // If a watch child is assigned the opposite value, the other
 // watch does not need to be moved.
  if ((g_nodes.value[first_watch] == flipValue(value)) ||
      (g_nodes.value[second_watch] == flipValue(value)))
    return SOLVER_OK;


 // Look at the first watch, if it is assigned 'value', move it
  if (g_nodes.value[first_watch] == value) {

   // Look for a new watch child in the list and swap it
    while (i < num_children) {

      child = getChildNode(parent, i);
      if (g_nodes.value[child] != value) {
        swapWatches(parent, 0, i);
        first_watch = child;
        if (g_nodes.value[child] == flipValue(value))
          return SOLVER_OK;
        break;
      }
      i++;

    }


   // If no watches are present, perform the appropriate propagation
   // rule on the OTHER watch
    if (i == num_children) {

      if (g_nodes.value[parent] == flipValue(value)) {
        return Solver_Assign(second_watch, flipValue(value), createNodeReason(PARENT_REASON, parent));
      } else if ((g_nodes.value[parent] == UNKNOWN) && (g_nodes.value[second_watch] == value)) {
        return Solver_Assign(parent, value, createNodeReason(CHILD_REASON, second_watch));
      }

    }

  }



 // Move the second watch if is also assigned 'value'
  if (g_nodes.value[second_watch] == value) {

   // Look for a new watch, note that i is not reinitialized, so we
   // do not perform redundant work if both watches need be moved.
    while (i < num_children) {

      child = getChildNode(parent, i);
      if (g_nodes.value[child] != value) {
        swapWatches(parent, 1, i);
        break;
      }
      i++;
    
    }


   // If no watches are present, perform the appropriate propagation
   // rule on the OTHER watch
    if (i == num_children) {

      if (g_nodes.value[parent] == flipValue(value)) {
        return Solver_Assign(first_watch, flipValue(value), createNodeReason(PARENT_REASON, parent));
      } else if ((g_nodes.value[parent] == UNKNOWN) && (g_nodes.value[first_watch] == value)) {
        return Solver_Assign(parent, value, createNodeReason(CHILD_REASON, second_watch));
      }

    }

  }


  return SOLVER_OK;
}








/******************************************************************
  PropagateFalseAnd
 ******************************************************************
 
  Description : Look through the children and assign the last child
                FALSE if all other children are TRUE
  Parameters  : node - the FALSE AND node to be processed
  Modifies    : May move watches or assign FALSE to a child
  Assumptions : The AND node to be processed must be FALSE
  Returns     : SOLVER_OK or SOLVER_CONFLICT (from Solver_Assign)

 ******************************************************************/
int propagateFalseAnd(int node) {

  int watch1 = getChildNode(node, 0);
  int watch2 = getChildNode(node, 0);

 // If some watch children are TRUE and they have been assigned at
 // the current level, attempt to move them
  if ( ((g_nodes.value[watch1] == TRUE) && (g_nodes.decision_level[watch1] == g_decision_level))
    || ((g_nodes.value[watch2] == TRUE) && (g_nodes.decision_level[watch2] == g_decision_level)) ) {

    return moveChildrenWatches(node, TRUE);


 // Otherwise, we may have already attempted to move them and there
 // are no free watches
  } else {

   // If one of the watches is TRUE, assign FALSE to the other watch
    assert(g_nodes.value[node] == FALSE);
    if ((g_nodes.value[watch1] == UNKNOWN) && (g_nodes.value[watch2] == TRUE)) {
      return Solver_Assign(watch1, FALSE, createNodeReason(PARENT_REASON, node));
    } else if ((g_nodes.value[watch1] == TRUE) && (g_nodes.value[watch2] == UNKNOWN)) {
      return Solver_Assign(watch2, FALSE, createNodeReason(PARENT_REASON, node));
    }
    
  }

  return SOLVER_OK;
}







/******************************************************************
  PropagateTrueOr
 ******************************************************************
 
  Description : Look through the children and assign the last child
                TRUE if all other children are FALSE
  Parameters  : node - the TRUE OR node to be processed
  Modifies    : May move watches or assign TRUE to a child
  Assumptions : The OR node to be processed must be TRUE
  Returns     : SOLVER_OK or SOLVER_CONFLICT (from an assignment)

 ******************************************************************/
int propagateTrueOr(int node) {

  int watch1 = getChildNode(node, 0);
  int watch2 = getChildNode(node, 0);

 // If some watch children are FALSE and they have been assigned at
 // the current level, attempt to move them
  if ( ((g_nodes.value[watch1] == FALSE) && (g_nodes.decision_level[watch1] == g_decision_level))
    || ((g_nodes.value[watch2] == FALSE) && (g_nodes.decision_level[watch2] == g_decision_level)) ) {

    return moveChildrenWatches(node, FALSE);


 // Otherwise, we may have already attempted to move them and there
 // are no free watches
  } else {

   // If one of the watches is FALSE, assign TRUE to the other watch
    assert(g_nodes.value[node] == TRUE);
    if ((g_nodes.value[watch1] == UNKNOWN) && (g_nodes.value[watch2] == FALSE)) {
      return Solver_Assign(watch1, TRUE, createNodeReason(PARENT_REASON, node));
    } else if ((g_nodes.value[watch1] == FALSE) && (g_nodes.value[watch2] == UNKNOWN)) {
      return Solver_Assign(watch2, TRUE, createNodeReason(PARENT_REASON, node));
    }
    
  }

  return SOLVER_OK;
}







/******************************************************************
  PropagateTrueChild
 ******************************************************************
 
  Description : Propagate the value of a TRUE node to the children
                (automatically for NOT and OR nodes, watches for
                AND nodes) and move the watches for every AND
                parent being watched by this child.
  Parameters  : child - the node that is assigned TRUE
  Modifies    : May assign values to the children or parents
  Assumptions : All the watch structures are initialized
  Returns     : SOLVER_OK or SOLVER_CONFLICT

 ******************************************************************/
int inline propagateTrueChild(int child) {

  int i, retval = SOLVER_OK, parent, num_parents = getNumParents(child);
  int info, prev_info;


 // Propagate to children if the node is a NOT or an OR
  for (i=0; i<num_parents; i++) {

    parent = getParentNode(child, i);
    switch (g_nodes.type[parent]) {

      case NOT_TYPE:
        retval = Solver_Assign(parent, FALSE, createNodeReason(CHILD_REASON, child));
        break;

      case OR_TYPE:
        retval = Solver_Assign(parent, TRUE, createNodeReason(CHILD_REASON, child));
        if (g_args.dcpropagation == TRUE)
          g_dcstack->push(parent);
        break;

      case AND_TYPE:
        break;

      default: // IFF and XOR calls go here
        assert(NULL);
    }

    if (retval != SOLVER_OK)
      return retval;
  }



 // Look through the AND watch list to move and TRUE watch children
  i = 0;
  info = (int) g_nodes.and_watchlist[child]->getElementAt(i);
  while (i < g_nodes.and_watchlist[child]->getSize()) {
    prev_info = info;

    retval = moveChildrenWatches(info, TRUE);
    if (retval != SOLVER_OK)
      return retval;

   // If the watch was successfully moved, the node would have been
   // removed from the vector and there is no need to incr the pos
    info = (int) g_nodes.and_watchlist[child]->getElementAt(i);
    if (info == prev_info)
      info = (int) g_nodes.and_watchlist[child]->getElementAt(++i);
  }


  return SOLVER_OK;
}






/******************************************************************
  PropagateFalseChild
 ******************************************************************

  Description : Propagate the value of a FALSE node to the children
                (automatically for NOT and AND nodes, watches for
                OR nodes) and move the watches for every OR
                parent being watched by this child.
  Parameters  : child - the node that is assigned FALSE
  Modifies    : May assign values to the children or parents
  Assumptions : All the watch structures are initialized
  Returns     : SOLVER_OK or SOLVER_CONFLICT
 
 ******************************************************************/
int inline propagateFalseChild(int child) {

  int i, retval = SOLVER_OK, parent, num_parents = getNumParents(child);
  int info, prev_info;


 // Propagate to children if the node is a NOT or an AND
  for (i=0; i<num_parents; i++) {

    parent = getParentNode(child, i);
    switch (g_nodes.type[parent]) {

      case NOT_TYPE:
        retval = Solver_Assign(parent, TRUE, createNodeReason(CHILD_REASON, child));
        break;

      case AND_TYPE:
        retval = Solver_Assign(parent, FALSE, createNodeReason(CHILD_REASON, child));
        if (g_args.dcpropagation == TRUE)
          g_dcstack->push(parent);
        break;

      case OR_TYPE: // Ignore OR nodes (dealt in the watches below)
        break;

      default: // IFF and XOR rules go here
        assert(NULL);
    }

    if (retval != SOLVER_OK)
      return retval;
  }



 // Look through the OR watch list to move and FALSE watch children
  i = 0;
  info = (int) g_nodes.or_watchlist[child]->getElementAt(i);
  while (i < g_nodes.or_watchlist[child]->getSize()) {
    prev_info = info;

    retval = moveChildrenWatches(info, FALSE);
    if (retval != SOLVER_OK)
      return retval;

   // If the watch was successfully moved, the node would have been
   // removed from the vector and there is no need to incr the pos
    info = (int) g_nodes.or_watchlist[child]->getElementAt(i);
    if (info == prev_info)
      info = (int) g_nodes.or_watchlist[child]->getElementAt(++i);
  }


  return SOLVER_OK;
}







/******************************************************************
  PropagateAND
 ******************************************************************
 
  Description : Propagate the value to the children of an AND node
                (some work is delegated to propagateFalseAND)
  Parameters  : node - the AND node propagated
  Modifies    : May assign TRUE to the children
  Assumptions : Same as the functions that are being called
  Returns     : SOLVER_OK or SOLVER_CONFLICT

 ******************************************************************/
int inline propagateAnd(int node) {

  int i, num_children, retval;

 // Assign TRUE to all the children
  if (g_nodes.value[node] == TRUE) {

    num_children = getNumChildren(node);
    for (i=0; i<num_children; i++) {
      retval = Solver_Assign(getChildNode(node, i), TRUE,
		      createNodeReason(PARENT_REASON, node));
      if (retval != SOLVER_OK)
        return retval;
    }
    return SOLVER_OK;


  } else {
    return propagateFalseAnd(node);
  }

}





/******************************************************************
  PropagateOR
 ******************************************************************
 
  Description : Propagate the value to the children of an OR node
                (some work is delegated to propagateFalseOR)
  Parameters  : node - the OR node propagated
  Modifies    : May assign FALSE to the children
  Assumptions : Same as the functions that are being called
  Returns     : SOLVER_OK or SOLVER_CONFLICT
  Notes       : This function can easily be combined with
                propagateAnd()

 ******************************************************************/
int inline propagateOr(int node) {

  int i, num_children, retval;

 // Assign FALSE to all the children
  if (g_nodes.value[node] == FALSE) {

    num_children = getNumChildren(node);
    for (i=0; i<num_children; i++) {
      retval = Solver_Assign(getChildNode(node, i), FALSE,
		      createNodeReason(PARENT_REASON, node));
      if (retval != SOLVER_OK)
        return retval;
    }
    return SOLVER_OK;


  } else {
    return propagateTrueOr(node);
  }

}






/******************************************************************
  PropagateNOT
 ******************************************************************
 
  Description : Assign the opposite value to the child
  Parameters  : node - the NOT node
  Modifies    : Assigns a value to the child
  Assumptions : <none>
  Returns     : SOLVER_OK or SOLVER_CONFLICT (output of Solver_Assign)

 ******************************************************************/
int inline propagateNot(int node) {

  return Solver_Assign(getChildNode(node,0), flipValue(g_nodes.value[node]),
                       createNodeReason(PARENT_REASON, node));

}







/******************************************************************
  PropagateNode
 ******************************************************************
 
  Description : Find out the type of the node and propagate the
                children values (see appropriate functions)
  Parameters  : node - index to the node
  Modifies    : <none>
  Assumptions : node is assigned the value TRUE or FALSE
  Returns     : SOLVER_OK or SOLVER_CONFLICT if propagation causes
                a conflict

 ******************************************************************/
int propagateNode(int node) {

  int retval = SOLVER_OK;


  assert((g_nodes.value[node] == TRUE) ||
         (g_nodes.value[node] == FALSE));


 // Process the node's children
  switch (g_nodes.type[node]) {

    case AND_TYPE:  retval = propagateAnd(node);  break;
    case NOT_TYPE:  retval = propagateNot(node);  break;
    case OR_TYPE:   retval = propagateOr(node);   break;
    case ATOM_TYPE: break;

    default: // IFF and XOR node types go here
      printf("ERROR: Propagating a node of unidentified type (%d)\n", g_nodes.type[node]);
      assert(NULL);
  }
  if (retval != SOLVER_OK)
    return retval;



 // Propagate any value to the parents
  if (g_nodes.value[node] == TRUE)
    return propagateTrueChild(node);
  else
    return propagateFalseChild(node);

}












/******************************************************************
  PropagateDCs
 ******************************************************************
 
  Description : Looks at each node on the stack that can no longer
                be a DC watch and attempts to move the watches to
                other nodes
  Parameters  : <none>
  Modifies    : empties the dc propagation stack
  Assumptions : structures are correctly allocated
  Returns     : SOLVER_OK always

 ******************************************************************/
int propagateDCs() {

 // Process each watch node on the stack
  int node;

  while (!g_dcstack->isEmpty()) {
    node = g_dcstack->pop();
    moveDCWatch(node);
  }

  return SOLVER_OK;
}







/******************************************************************
  FindChildWithValue
 ******************************************************************
 
  Description : Looks for a child that has a given value
  Parameters  : parent - index to the parent node
                value - TRUE or FALSE depending on what value we
                        are looking for
  Modifies    : <none>
  Assumptions : <none>
  Returns     : the index of the child node with the given value
                or -1 if no such children exists

 ******************************************************************/
int inline findChildWithValue(int parent, short value) {

  int i = 0, node, num_children = getNumChildren(parent);
  
 // Look through all the children of the Don't Care
  while (i < num_children) {
    node = getChildNode(parent, i);
    if (g_nodes.value[node] == value)
      return node;
    i++;
  }
  
  return -1;
}











/******************************************************************
  FindDCWatch
 ******************************************************************
 
  Description : Find a new watch parent for a node
  Parameters  : watched_node - index of the node that no longer has
                               a valid watch parent
  Modifies    : <none>
  Assumptions : <none>
  Returns     : the index of the new watch or -1 if no new watch
                is found

 ******************************************************************/
int inline findDCWatch(int watched_node) {

  int i = 0, node, num_parents = getNumParents(watched_node);


 // Look for an UNKNOWN parent or a FALSE AND parent with no FALSE
 // children or a TRUE OR parent with no TRUE children
  while (i < num_parents) {
    node = getParentNode(watched_node, i++);
    if ( (g_nodes.value[node] == UNKNOWN) ||
        ((g_nodes.value[node] == TRUE) && (g_nodes.type[node] == OR_TYPE)
          && (findChildWithValue(node, TRUE) == -1)) ||
        ((g_nodes.value[node] == FALSE) && (g_nodes.type[node] == AND_TYPE)
          && (findChildWithValue(node, FALSE) == -1)) )
      return node;
  }

  return -1;
}










/******************************************************************
  MoveDCWatch
 ******************************************************************
 
  Description : Looks for a new watch and either assigns DONT_CARE
                to the child or moves the watch
  Parameters  : node - index to the parent node that forces its
                       children to be Don't Cares
  Modifies    : may move some watches or assign values to nodes
  Assumptions : this function is not called if g_args.dcpropagation is
                FALSE
  Returns     : <none>

 ******************************************************************/
void moveDCWatch(int node) {

  int i = 0, child, dummy, new_watch;
  reasoninfo	reason = {0,{0}};

 // Look through the dc list
  while (i < g_nodes.dc_watchlist[node]->getSize()) {

   // Move the watches for UNKNOWN children (else ignore)
    child = g_nodes.dc_watchlist[node]->getElementAt(i);
    if (g_nodes.value[child] == UNKNOWN) {

      new_watch = findDCWatch(child);
      if (new_watch == -1) {
        Undo_Assign(child, DONT_CARE, reason);
        g_dcstack->push(child);
        i++;
      } else {
        g_nodes.dc_watchlist[node]->remove(i, &dummy);
        g_nodes.dc_watchlist[new_watch]->add(child);
      }

    } else {
      i++;
    }

  }

}










/******************************************************************
  PropagateNoGoods
 ******************************************************************
 
  Description : Propagates any assignments that are forced because
                of nogoods
  Parameters  : node - index of the node that is being propagated
  Modifies    : may change some watches for nodes or assign values
                to the nodes
  Assumptions : as usual, structures are all allocated and nogoods
                are setup correctly (with a mark for entries that
                are watches)
  Returns     : SOLVER_OK or SOLVER_CONFLICT if propagation yields
                a conflict

 ******************************************************************/
int propagateNoGoods(int node) {

  Vector *v = (g_nodes.value[node] == TRUE)?
         g_nodes.true_watchlist[node]:g_nodes.false_watchlist[node];
  int *info, *prev_info, i, retval;


 // Look through the watch list
  i = 0;
  info = (int*) v->getElementAt(i);
  while (i < v->getSize()) {
    prev_info = info;

    retval = moveNoGoodWatch(node, info);
    if (retval != SOLVER_OK)
      return retval;

    info = (int*) v->getElementAt(i);
    if (info == prev_info)
      info = (int*) v->getElementAt(++i);
  }

  return SOLVER_OK;
}












/******************************************************************
  FindNewWatches
 ******************************************************************
 
  Description : Find the best watches for a given nogood
  Parameters  : nogood - list of literals (terminated with 0)
                watch1 - pointer to the best watch to be found
                watch2 - pointer to the second best watch
  Modifies    : Makes watch1 and watch2 contain the new watches
  Assumptions : The current watch literals are the first two in the
                nogood.
  Returns     : <none>

 ******************************************************************/
void inline findNewWatches(int *nogood, int *watch1, int *watch2) {

 // Variables for two unknowns (which are equivalent) and the two
 // 'best' watches which are assigned the opposite value at the
 // highest decision level
  int unknown1 = -1, unknown2 = -1, level1 = 0, level2 = 0, pos = 2;
  int good_watch1 = -1, good_watch2 = -1;
  int node, value, level;


 // Look for a new watch literal (starting at position 2)
  while (nogood[pos] != 0) {

   // Get the info for the current literal
    node = nogood[pos] >> 1;
    value = (nogood[pos] & 0x00000001)?TRUE:FALSE;
    level = g_nodes.decision_level[node];


   // If the node is currently assigned the opposite value
    if (g_nodes.value[node] == flipValue(value)) {

     // See if it is better than the current 'best' watches
      if ((good_watch1 == -1) || (level < level1)) {
        good_watch2 = good_watch1;
        good_watch1 = pos;
        level2 = level1;
        level1 = level;
      } else if ((good_watch2 == -1) || (level < level2)) {
        good_watch2 = pos;
        level2 = level;
      }

   // Otherwise, if a DC or UNKNOWN, it can also be a watch
    } else if (g_nodes.value[node] != value) {

      if ((unknown1 == -1) || (node < (nogood[unknown1] >> 1))) {
        unknown2 = unknown1;
        unknown1 = pos;
      } else if (unknown2 == -1) {
        unknown2 = pos;
      }

    }

    pos++;
  }


 // Set the two 'best' watches from the available info
  if (good_watch1 < 0) {
    good_watch1 = unknown1;
    good_watch2 = unknown2;
  } else if (good_watch2 < 0) {
    good_watch2 = unknown1;
  }


  *watch1 = good_watch1;
  *watch2 = good_watch2;
}








/******************************************************************
  MoveWatch
 ******************************************************************
 
  Description : Swap the watch node in a nogood since the watches
                must be the first two nodes of a nogood
  Parameters  : nogood  - the list of literals in the nogood
                old_loc - location of the moved watch (0 or 1)
                new_loc - location of the new literal to move
  Modifies    : Moves literals inside the nogood and removes/adds
                the nogood to the watchlist vector
  Assumptions : The two integers PRECEEDING the nogood contain the
                indices of the current watches (where this nogood
                is located in their respective watch vectors)
  Returns     : <none>

 ******************************************************************/
void inline moveWatch(int *nogood, int old_loc, int new_loc) {

 // Find the index of the watch and remove all former watch info
  int tmp, *loc = (&nogood[old_loc]) - 2;
  NoGood_RemoveWatch(nogood, old_loc);

 // Add new watch info for the new watch
  if (nogood[new_loc] & 0x00000001) {
    *loc = g_nodes.true_watchlist[ nogood[new_loc] >> 1 ]->add((int)nogood) | 0x80000000;
  } else {
    *loc = g_nodes.false_watchlist[ nogood[new_loc] >> 1 ]->add((int)nogood) | 0x80000000;
  }

 // Swap the watch to be in position 0 or 1 in the nogood
  tmp = nogood[old_loc];
  nogood[old_loc] = nogood[new_loc];
  nogood[new_loc] = tmp;

}










/******************************************************************
  MoveNoGoodWatch
 ******************************************************************
 
  Description : Move the watch to an unassigned entry.  If there
                are no unassigned entries, then try to assign the
                opposite value to the other watch
  Parameters  : w - pointer to the location of the current watch
                first - pointer to the location of the first watch
                        in the current list (used in order to
	    remove w from the list)
  Modifies    : if the watch is moved, w will be set to NULL
  Assumptions : the watches in the nogood are marked and the nogood
                begins and ends with 0s
  Returns     : SOLVER_OK or SOLVER_CONFLICT if the other watch
                is assigned

 ******************************************************************/
int moveNoGoodWatch(int index, int *nogood) {

  int good_watch1 = -1, good_watch2 = -1;


 // Removing this IF can sometimes improve performance... (strange)
  if ((g_nodes.value[nogood[0] >> 1] == flipValue(nogood[0] & 0x00000001) ) ||
      (g_nodes.value[nogood[1] >> 1] == flipValue(nogood[1] & 0x00000001) )) {
    return SOLVER_OK;
  }


 // Find the best watches for the nogood
  findNewWatches(nogood, &good_watch1, &good_watch2);


 // Move the first watch if necessary
  if (g_nodes.value[nogood[0] >> 1] == ( nogood[0] & 0x00000001 )) {

    if (good_watch1 < 0) {
      return Solver_Assign(nogood[1] >> 1,
                         flipValue(nogood[1] & 0x00000001),
                         createNoGoodReason(NOGOOD_REASON, nogood));
    } else {
      moveWatch(nogood, 0, good_watch1);
      good_watch1 = good_watch2;
    }
  }


 // Move the second watch if necessary
  if (g_nodes.value[nogood[1] >> 1] == ( nogood[1] & 0x00000001 )) {

    if (good_watch1 < 0) {
      return Solver_Assign(nogood[0] >> 1,
                         flipValue(nogood[0] & 0x00000001),
                         createNoGoodReason(NOGOOD_REASON, nogood));
    } else {
      moveWatch(nogood, 1, good_watch1);
    }
  }


  return SOLVER_OK;
}


















/* These functions are incomplete and not fully tested as we don't
   have many problems involving IFFs and XORs */




int propagateIffNode(int node) {

  int /*count_true = 0,*/ count_false = 0, count_unknown = 0;
  int i, num_children = getNumChildren(node);
  int child, unknown = -1;


 /* Find the number of TRUE and FALSE children */
  for (i=0; i<num_children; i++) {
    child = getChildNode(node, i);
    switch (g_nodes.value[child]) {
      case TRUE:     /* count_true++; */   break;
      case FALSE:     count_false++;   break;
      case UNKNOWN:   if (count_unknown != 0) return SOLVER_OK;
                      count_unknown++; unknown = child; break;
      case DONT_CARE: if (count_unknown != 0) return SOLVER_OK;
                      count_unknown++; unknown = child; break;
      default:        printf("Unknown value\n");  assert(NULL);
    }
  }


 /* If there are no unknowns, verify the count */
  if (count_unknown == 0) {
    if (count_false % 2 == 0)
      return Solver_Assign(node, TRUE,
        createNodeReason(CHILD_REASON, getChildNode(node, 1)));
    else
      return Solver_Assign(node, FALSE,
        createNodeReason(CHILD_REASON, getChildNode(node, 1)));
  }


 /* Set the only unknown child */
  if (count_unknown == 1) {
    if (((g_nodes.value[node] == TRUE)  && (count_false % 2 == 0)) ||
        ((g_nodes.value[node] == FALSE) && (count_false % 2 == 1))) {
      return Solver_Assign(unknown, TRUE,
                           createNodeReason(PARENT_REASON, node));
    } else {
      return Solver_Assign(unknown, FALSE,
                           createNodeReason(PARENT_REASON, node));
    }
  }

  return SOLVER_OK;
}






int propagateXorNode(int node) {

  int count_true = 0,/* count_false = 0,*/ count_unknown = 0;
  int i, num_children = getNumChildren(node);
  int child, unknown = -1;


 /* Find the number of TRUE and FALSE children */
  for (i=0; i<num_children; i++) {
    child = getChildNode(node, i);
    switch (g_nodes.value[child]) {
      case TRUE:      count_true++;    break;
      case FALSE:    /* count_false++; */  break;
      case UNKNOWN:   if (count_unknown != 0) return SOLVER_OK;
                      count_unknown++; unknown = child; break;
      case DONT_CARE: if (count_unknown != 0) return SOLVER_OK;
                      count_unknown++; unknown = child; break;
      default:        printf("Unknown value\n");  assert(NULL);
    }
  }


 /* If there are no unknowns, verify the count */
  if (count_unknown == 0) {
    if (count_true % 2 == 0)
      return Solver_Assign(node, FALSE,
        createNodeReason(CHILD_REASON, getChildNode(node, 1)));
    else
      return Solver_Assign(node, TRUE,
        createNodeReason(CHILD_REASON, getChildNode(node, 1)));
  }


 /* Set the only unknown child */
  if (count_unknown == 1) {
    if (((g_nodes.value[node] == TRUE)  && (count_true % 2 == 0)) ||
        ((g_nodes.value[node] == FALSE) && (count_true % 2 == 1))) {
      return Solver_Assign(unknown, TRUE,
                           createNodeReason(PARENT_REASON, node));
    } else {
      return Solver_Assign(unknown, FALSE,
                           createNodeReason(PARENT_REASON, node));
    }
  }

  return SOLVER_OK;
}





int propagateIffParent(int parent, int child) {

 /* int count_true = 0; */
  int count_false = 0;
  int i, node, num_children = getNumChildren(parent);


 /* Find the number of TRUE and FALSE children */
  for (i=0; i<num_children; i++) {
    node = getChildNode(parent, i);
    switch (g_nodes.value[node]) {
      case TRUE:     /* count_true++; */   break;
      case FALSE:     count_false++;   break;
      case UNKNOWN:   return SOLVER_OK;
      case DONT_CARE: return SOLVER_OK;
      default:        printf("Unknown value\n");  assert(NULL);
    }
  }


 /* Set the value if all nodes are set */
  if (count_false % 2 == 0) {
    return Solver_Assign(parent, TRUE,
                         createNodeReason(CHILD_REASON, child));
  } else {
    return Solver_Assign(parent, FALSE,
                         createNodeReason(CHILD_REASON, child));
  }
}





int propagateXorParent(int parent, int child) {

  int count_true = 0;
 /* int count_false = 0; */
  int i, node, num_children = getNumChildren(parent);


 /* Find the number of TRUE and FALSE children */
  for (i=0; i<num_children; i++) {
    node = getChildNode(parent, i);
    switch (g_nodes.value[node]) {
      case TRUE:      count_true++;    break;
      case FALSE:     /* count_false++; */  break;
      case UNKNOWN:   return SOLVER_OK;
      case DONT_CARE: return SOLVER_OK;
      default:        printf("Unknown value\n");  assert(NULL);
    }
  }


 /* Set the value if all nodes are set */
  if (count_true % 2 == 1) {
    return Solver_Assign(parent, TRUE,
                         createNodeReason(CHILD_REASON, child));
  } else {
    return Solver_Assign(parent, FALSE,
                         createNodeReason(CHILD_REASON, child));
  }
}






