/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  October 19th 2003 (April 2004)

  *****************************************************************

  Solver.cpp

  Contains the functions required to solve an expression represented
  compactly in a contiguous block of memory and indexed globally by
  by g_nodes.

  *****************************************************************/

#include "solver.h"
#include "stats.h"

#include "undo.h"
#include "propagate.h"
#include "nogood.h"





/********************** Global Variables **************************/

 // Global reference to the last node that created a conflict
int g_conflict_node = -1;
reasoninfo g_conflict_reason = {0,{0}};
int g_nogoods_remaining = 0;

 // Array of nodes sorted by score (updated dynamically)
node_ordering *g_ordered_nodes = NULL;
int g_choice_index = 0;      // Last node chosen



/******************* Local Function Prototypes ********************/

int solve();
int choose();
int restore();








/***************** Externally Available Functions *****************/


/******************************************************************
  Solver_SolveDAG
 ******************************************************************
 
  Description : Try to find a satisfiable assignment to the problem
                after assignment TRUE to the root
  Parameters  : <none>
  Modifies    : Allocates memory required to solve the problem
                and deallocates it
  Assumes     : g_nodes indexes a properly parsed DAG,
                the root is indexed by node 0
  Returns     : SOLVER_OK if all went well
                (currently the only return value possible)

 ******************************************************************/
int Solver_SolveDAG() {

  int i;

 // Make sure the DAG has been built
  assert(g_num_nodes > 0);


 // Allocate memory used for solving
  g_ordered_nodes = (node_ordering*) malloc
                          ((g_num_nodes-1) * sizeof(node_ordering));
  assert(g_ordered_nodes);
  for (i=1; i<g_num_nodes; i++) {
    g_ordered_nodes[i-1].node = i;
    g_ordered_nodes[i-1].score = 0;
    Solver_SetTrueScore(i, 0);
    Solver_SetFalseScore(i, 0);
    g_nodes.score_index[i] = i-1;
  }

  g_choice_index = 0;
  g_nogoods_remaining = 0;
  Propagate_Initialize();
  Undo_Initialize();
  NoGood_Initialize();


 // Solve by assigning TRUE to the root
  Stats_StartTimer(STATS_SOLVER);
  if (solve() == SOLVER_SATISFIABLE)
    g_is_satisfiable = TRUE;
  else
    g_is_satisfiable = FALSE;
  Stats_StopTimer(STATS_SOLVER);


 // Deallocate memory used in solving
  NoGood_Free();
  Undo_Free();
  Propagate_Free();
  free(g_ordered_nodes);

  return SOLVER_OK;
}








/******************************************************************
  Solver_Assign
 ******************************************************************
 
  Description : Assign a value to a node, place it on the propagate
                stack and create an undo record
  Parameters  : node - index of the node that has a new value
                value - value to be assigned to this node
                reason - reason for the assignment
  Modifies    : changes the value of the node
  Assumptions : <none>
  Returns     : SOLVER_CONFLICT if there was a conflict, else
                SOLVER_OK

 ******************************************************************/
int Solver_Assign(int node, byte value, reasoninfo reason) {


 // Do nothing if the value is already assigned or a Don't Care
  if ((g_nodes.value[node] == value) || (g_nodes.value[node] == DONT_CARE))
    return SOLVER_OK;

 // Output the assignment and reason for debugging
  if (g_args.verbose >= 3) {
    int reason_node = getNodeReason(reason);
    printf("Assigning value %s to %s node %d (%s %d)\n",
           (value == TRUE)?"TRUE":"FALSE", printNodeType(node), node,
           (getReasonType(reason) == CHOICE_REASON)?"CHOICE":
           (getReasonType(reason) == NOGOOD_REASON)?"NOGOOD":
           (getReasonType(reason) == PARENT_REASON)?"PARENT":
           (getReasonType(reason) == CHILD_REASON)?"CHILD":
           (getReasonType(reason) == UIP_REASON)?"UIP":"UNKNOWN",
           reason_node);
  }
  assert((value == TRUE) || (value == FALSE));


 // Assign the new value
  if (g_nodes.value[node] == UNKNOWN) {

    Undo_Assign(node, value, reason);
    Propagate_Add(node);
    return SOLVER_OK;

  }


 // Create a conflict if the values do not match
  g_conflict_node = node;
  g_conflict_reason = reason;
  return SOLVER_CONFLICT;

}












/********************** Local Functions ***************************/



/******************************************************************
  Solve
 ******************************************************************
 
  Description : Sets the root to TRUE and loops through the main
                functions of the solver (propagate, restore, choose)
  Parameters  : <none>
  Modifies    : calls functions that modify various data structs
  Assumptions : all data structures are correctly initialized
  Returns     : SOLVER_UNSATISFIABLE or SOLVER_SATISFIABLE

 ******************************************************************/
int solve() {

 // Assign TRUE to the root
  int retval = Solver_Assign(0, TRUE, createNodeReason(CHOICE_REASON, 0));


  while (1) {

   // Propagate the assignment and make any possible inferences
    if (retval == SOLVER_OK) {
      retval = Propagate_Propagate();
    }


   // If we have a conflict, restore and learn a new nogood
    if ( (retval != SOLVER_OK) &&
         (restore() == SOLVER_UNSATISFIABLE) ) {
      return SOLVER_UNSATISFIABLE;
    }


   // Choose and make the next assignment
    g_stats.num_choices++;
    if ((g_args.verbose >= 2) && (g_stats.num_choices % 25000 == 0))
      printf("Making choice number %d (%d nogoods)\n",
             g_stats.num_choices, g_stats.nogoods_learned);
    retval = choose();
    if (retval == SOLVER_SATISFIABLE)
      return SOLVER_SATISFIABLE;

  }

}








/******************************************************************
  Choose
 ******************************************************************
 
  Description : Select the next unassigned variable according to the
                VSIDS score
  Parameters  : <none>
  Modifies    : May decay the VSIDS scores of the nodes
  Assumptions : The ordered_nodes array is sorted according to the
                overall score of the nodes.
  Returns     : The result of assigning the node (which should
                always be SOLVER_OK since the node must be initially
                unassigned)

 ******************************************************************/
int choose() {

 // Decay the scores every so often
  if (g_stats.num_choices % SOLVER_VDS_DECAY_RATE == SOLVER_VDS_DECAY_RATE-1)
    Solver_DecayVariableScores(SOLVER_VDS_DECAY_FACTOR);

 // Find the first unassigned variable (as ordered by score)
  int i = g_choice_index;
  bool avoided = FALSE;
  while ((i<g_num_nodes-1) && (g_nodes.value[g_ordered_nodes[i].node] != UNKNOWN)) {
    if (g_nodes.value[g_ordered_nodes[i].node] == DONT_CARE)
      avoided = TRUE;
    i++;
  }
  if (avoided)
    g_stats.num_avoided_choices++;


 // If there are no assignments, the problem is satisfiable
  if (i == g_num_nodes-1) {
    for (i=0; i<g_num_nodes; i++)
      assert(g_nodes.value[i] != UNKNOWN);
    return SOLVER_SATISFIABLE;
  }
  g_choice_index = i + 1;

/*
 // Randomize nodes of the same score
  g_choice_index = i; // i+1 if no randomness
  int j = i+1, count = 0;
  while (g_ordered_nodes[i].score == g_ordered_nodes[j].score) {
    if (g_nodes.value[g_ordered_nodes[j++].node] == UNKNOWN)
      count++;
  }
  if (count > 0) {
    count = rand() % count;
    for (j=0; j<count; j++) {
      while (g_nodes.value[g_ordered_nodes[++i].node] != UNKNOWN)
        ;
    }
  }
*/

 // Make the assignment
  int node = g_ordered_nodes[i].node;
  byte value = ((Solver_GetTrueScore(node) > Solver_GetFalseScore(node))?FALSE:TRUE);
  return Solver_Assign(node, value, createNodeReason(CHOICE_REASON, 0));
}
















/******************************************************************
  Restore
 ******************************************************************
 
  Description : This function unrolls the assignments up to the
                nearest safe point (non-conflict)
  Parameters  : <none>
  Modifies    : reverts nodes back to their previous assignment
                (which must be UNKNOWN)
  Assumptions : g_conflict_node and g_conflict_reason contain the
                appropriate information (as set by Solver_Assign)
  Returns     : SOLVER_OK or SOLVER_UNSATISFIABLE if we backtrack
                back to the root

 ******************************************************************/
int restore() {

  int retval, initial_level = g_decision_level;


 // Unroll the assignments as far as we can, note that unrolling
 // can yield more assignments which are propagated and can create
 // more conflicts.  Undo_Unroll returns SOLVER_OK or
 // SOLVER_UNSATISFIABLE.
  g_choice_index = 0;        // For VDS
  retval = Undo_Unroll();
  if (retval != SOLVER_OK) {
    return retval;
  }



 // Update the backtracking counters
  int backtrack_size = initial_level - g_decision_level;
  g_stats.num_backtracks++;
  g_stats.total_backtracks += backtrack_size;
  if (backtrack_size > g_stats.longest_backtrack)
    g_stats.longest_backtrack = backtrack_size;



 // Make sure the nogood database doesn't get too large
  if (g_num_tmp_nogoods - g_nogoods_remaining > NOGOOD_CLEANUP_COUNT) {
    NoGood_Cleanup();
    g_nogoods_remaining = g_num_tmp_nogoods;
  }


  return retval;
}













/************************* VDS Functions **************************/



/******************************************************************
  Solver_UpdateVariableScores
 ******************************************************************
 
  Description : Increment the score of each node in the nogood
  Parameters  : nogood - the learned nogood
                size - the number of entries in the nogood
  Modifies    : Changes the score of nodes
  Assumptions : nogood is non-NULL and has size entries
  Returns     : <none>

 ******************************************************************/
void Solver_UpdateVariableScores(int *nogood, int size) {

  int i, j, step, new_score;
  int node;
  byte value;


 // Look at each entry in the nogood
  for (i=0; i<size; i++) {

   // Convert the nogood entry to a node and value
    node = nogood[i] >> 1;
    value = (nogood[i]&0x00000001)?TRUE:FALSE;
    if (node == 0)
      continue;


   // Increment the node's score
    if (value == TRUE) {
      Solver_IncrementTrueScore(node);
    } else {
      Solver_IncrementFalseScore(node);
    }


   // Calculate the new score for this node
    new_score = Solver_GetScore(node);
    if (new_score <= g_ordered_nodes[g_nodes.score_index[node]].score)
      continue;
    assert(new_score == (g_ordered_nodes[g_nodes.score_index[node]].score + 1));

/*
   // Linear Search
   // Get the next spot (can do this more efficiently probably)
    j = g_nodes.score_index[node];
    while ((j > 0) && (g_ordered_nodes[j-1].score < new_score))
      j--;
*/

   // Binary search based on Zchaff's update
    step = SOLVER_VDS_INITIAL_STEP;
    for (j = g_nodes.score_index[node]; j >= 0; j -= step) {
      if (g_ordered_nodes[j].score >= new_score)
        break;
    }
    j += step;
    for (step = step>>1; step > 0; step = step>>1) {
      if (j - step >= 0) {
        if (g_ordered_nodes[j - step].score < new_score)
          j -= step;
      }
    }


   // If the node does not have to move, simply update the score
    if (j == g_nodes.score_index[node]) {
      g_ordered_nodes[j].score = new_score;
      continue;
    }


   // Swap the nodes 
    assert(g_ordered_nodes[j].score == g_ordered_nodes[g_nodes.score_index[node]].score);
    assert(g_nodes.score_index[g_ordered_nodes[j].node] == j);
    g_nodes.score_index[g_ordered_nodes[j].node] = g_nodes.score_index[node];
    g_ordered_nodes[g_nodes.score_index[node]].node = g_ordered_nodes[j].node;

    g_nodes.score_index[node] = j;
    g_ordered_nodes[j].score = new_score;
    g_ordered_nodes[j].node = node;

  }


}






/******************************************************************
  Solver_DecayVariableScores
 ******************************************************************
 
  Description : Divide the score of all the nodes by a constant
  Parameters  : factor - the decay constant (2^factor because of the
                         bit shifting used in this function)
  Modifies    : Changes the score of all the nodes
  Assumptions : As usual, all the data structures must be initialized
  Returns     : <none>

 ******************************************************************/
void Solver_DecayVariableScores(int factor) {

  int i;
  for (i=0; i<g_num_nodes-1; i++) {
    Solver_SetTrueScore( g_ordered_nodes[i].node, Solver_GetTrueScore(g_ordered_nodes[i].node) >> factor );
    Solver_SetFalseScore( g_ordered_nodes[i].node, Solver_GetFalseScore(g_ordered_nodes[i].node) >> factor );
    g_ordered_nodes[i].score >>= factor;
  }

}











