/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  October 18th 2003 (April 2004)

  *****************************************************************

  Stats.cpp

  Contains the functions that are used to time regions of the dpll
  solver.  This file also initializes the global stat counters and
  prints a summary of the statistics accumulated.

  *****************************************************************/

#include "stats.h"





/******************* Local Function Prototypes ********************/
#ifdef WIN32
static double GetWinRunTime();
#endif
double getInternalRunTime();





/*************** Global Variables (in this module) ****************/
#ifdef WIN32
int g_isNT;
#endif

double g_parsing_time, g_tmp_parsing_time;
double g_merging_time, g_tmp_merging_time;
double g_compact_time, g_tmp_compact_time;
double g_solver_time, g_tmp_solver_time;
double g_debug1_time, g_tmp_debug1_time;
double g_debug2_time, g_tmp_debug2_time;
double g_debug3_time, g_tmp_debug3_time;





/********************** External Functions ************************/



/******************************************************************
  Stats_Initialize
 ******************************************************************
 
  Description : Initializes all the counters that will be used in
                this problem
  Parameters  : <none>
  Modifies    : sets all the global counters to zero
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
void Stats_Initialize() {

#ifdef WIN32
  /* Detect platform */
  {
    OSVERSIONINFO osvi; 
    osvi.dwOSVersionInfoSize=sizeof(OSVERSIONINFO); 
    GetVersionEx(&osvi); 
    if(osvi.dwPlatformId==VER_PLATFORM_WIN32_NT)
      g_isNT=TRUE;
    else
      g_isNT=FALSE;
  } 
#endif /* WIN32 */


 // Stats Counters
  g_stats.num_choices = 0;
  g_stats.num_avoided_choices = 0;
  g_stats.num_implications = 0;
  g_stats.num_dc_implications = 0;
  g_stats.nogoods_learned = 0;
  g_stats.literals_learned = 0;
  g_stats.num_backtracks = 0;
  g_stats.total_backtracks = 0;
  g_stats.longest_backtrack = 0;
  g_stats.num_nogoods_examined = 0;
  g_stats.num_literals_examined = 0;
  g_stats.num_redundant_literals = 0;
  g_stats.num_restarts = 0;


 // Initialize timers
  g_parsing_time = 0.0;
  g_merging_time = 0.0;
  g_compact_time = 0.0;
  g_solver_time = 0.0;
  
  g_debug1_time = 0.0;
  g_debug2_time = 0.0;
  g_debug3_time = 0.0;

}







/******************************************************************
  Stats_PrintResults
 ******************************************************************
 
  Description : Prints a summary of the global information
  Parameters  : <none>
  Modifies    : <none>
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
void Stats_PrintResults() {

  printf("\nSummary (%s%s):\n", (g_is_satisfiable == TRUE)?"Satisfiable":(g_is_satisfiable == FALSE)?"Unsatisfiable":"Unknown", (g_is_satisfiable)?((g_is_verified)?" - Verified":" - Inconsistent"):"");
  printf("  %d nodes (%d atoms) in input file (initially %d)\n", g_num_nodes, g_desc.num_atoms, g_desc.max_index);
  printf("  (%d max children, %d max parents, %.2f avg)\n", g_desc.max_children, g_desc.max_parents, (((double)g_desc.num_total_children)/g_num_nodes));
  printf("  Parsing took %.2f cpu secs, Merging took %.2f cpu secs\n", (g_parsing_time>0)?g_parsing_time:0.0, (g_merging_time>0)?g_merging_time:0.0);
  printf("  Converting to a compact representation took %.2f cpu secs\n", (g_compact_time>0)?g_compact_time:0.0);
  printf("\n");
  printf("  %d decisions (%d avoided), %d nogoods (%lld literals)\n", g_stats.num_choices, g_stats.num_avoided_choices, g_stats.nogoods_learned, g_stats.literals_learned);
  printf("  %lld implications, %lld Don't Care implications (%.0f total / sec)\n", g_stats.num_implications, g_stats.num_dc_implications, (g_solver_time>0)?(((double)(g_stats.num_implications + g_stats.num_dc_implications))/g_solver_time):0.0);
  if (g_stats.num_nogoods_examined != 0) printf("  %d reductions, %lld literals, %lld redundant (%.2f of %.2f)\n", g_stats.num_nogoods_examined, g_stats.num_literals_examined, g_stats.num_redundant_literals, (g_stats.num_nogoods_examined == 0)?0.0:(((double)g_stats.num_redundant_literals)/g_stats.num_nogoods_examined), (g_stats.num_nogoods_examined == 0)?0.0:(((double)g_stats.num_literals_examined)/g_stats.num_nogoods_examined));
  printf("  %d backtracks, %.2f average length (longest is %d), %d restarts\n", g_stats.num_backtracks, (g_stats.num_backtracks == 0)?0.0:((double)g_stats.total_backtracks)/g_stats.num_backtracks, g_stats.longest_backtrack, g_stats.num_restarts);
  printf("  Solving problem took %.3f cpu secs\n", (g_solver_time>0)?g_solver_time:0.0);
  if (g_debug1_time > 0.0001) printf("  Debug1 time is %.3f cpu secs\n", g_debug1_time);
  if (g_debug2_time > 0.0001) printf("  Debug2 time is %.3f cpu secs\n", g_debug2_time);
  if (g_debug3_time > 0.0001) printf("  Debug3 time is %.3f cpu secs\n", g_debug3_time);
  printf("\n\n");

}






/******************************************************************
  Stats_PrintLine
 ******************************************************************
 
  Description : Print a summary line for input to a spreadsheet
  Parameters  : filename - the input file
  Modifies    : <none>
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
void Stats_PrintLine(char *filename) {

  printf("%s\t%.2f\t%d\t%lld\t%lld\t%d\t%.2f\t%d\t%d\t%lld\t%.2f\t%d\t%lld\t%lld\t%.2f\t%.2f\t%s\t%s\n",
         filename,
         (g_solver_time>0)?g_solver_time:0.0,
         g_stats.num_choices,
         g_stats.num_implications,
         g_stats.num_dc_implications,
         g_stats.num_backtracks,
         (g_stats.num_backtracks == 0)?0.0:((double)g_stats.total_backtracks)/g_stats.num_backtracks,
         g_stats.num_restarts,
         g_stats.nogoods_learned,
         g_stats.literals_learned,
         (g_stats.nogoods_learned > 0)?((double)g_stats.literals_learned)/g_stats.nogoods_learned:0.0,
         g_stats.num_nogoods_examined,
         g_stats.num_literals_examined,
         g_stats.num_redundant_literals,
         (g_stats.num_nogoods_examined == 0)?0.0:(((double)g_stats.num_redundant_literals)/g_stats.num_nogoods_examined),
         (g_stats.num_nogoods_examined == 0)?0.0:(((double)g_stats.num_literals_examined)/g_stats.num_nogoods_examined),
         (g_is_satisfiable)?"SAT":"UNSAT",
         (g_is_verified)?"TRUE":"FALSE");

}



/******************************************************************
  Stats_PrintParseLine
 ******************************************************************
 
  Description : Print a summary line of the problem structure
  Parameters  : filename - the input file
  Modifies    : <none>
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
void Stats_PrintParseLine(char *filename) {

//  printf("Filename\tParse Time\tMerge Time\tCompact Time\tNodes\tAtoms\tMax Children\tMax Parents\tAvg Children\tAvg Parents\tANDs\tORs\tNOTs\tAvg AND Children\tAvg OR Children\tMax AND Children\tMax OR Children\n");
  printf("%s\t%.2f\t%.2f\t%.2f\t%d\t%d\t%d\t%d\t%.2f\t%.2f\t%d\t%d\t%d\t%.2f\t%.2f\t%d\t%d\n",
         filename, (g_parsing_time>0)?g_parsing_time:0.0, (g_merging_time>0)?g_merging_time:0.0, (g_compact_time>0)?g_compact_time:0.0,
         g_num_nodes, g_desc.num_atoms, g_desc.max_children, g_desc.max_parents, (((double)g_desc.num_total_children)/g_num_nodes),
         (((double)g_desc.num_total_children)/g_num_nodes), g_desc.num_and_nodes, g_desc.num_or_nodes, g_desc.num_not_nodes,
         (((double)g_desc.num_and_children)/g_desc.num_and_nodes), (((double)g_desc.num_or_children)/g_desc.num_or_nodes),
         g_desc.max_and_children, g_desc.max_or_children);
}



/******************************************************************
  Stats_StartTimer
 ******************************************************************
 
  Description : Gets the time for the start of a portion of the
                solver
  Parameters  : timed_region - integer representing the region that
                               is to be timed
  Modifies    : <none>
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
void Stats_StartTimer(int timed_region) {
  double tmp_time = getInternalRunTime();
  switch (timed_region) {
    case STATS_PARSING:  g_tmp_parsing_time = tmp_time;  break;
    case STATS_MERGING:  g_tmp_merging_time = tmp_time;  break;
    case STATS_COMPACT:  g_tmp_compact_time = tmp_time;  break;
    case STATS_SOLVER:   g_tmp_solver_time = tmp_time;   break;
    case STATS_DEBUG1:   g_tmp_debug1_time = tmp_time;   break;
    case STATS_DEBUG2:   g_tmp_debug2_time = tmp_time;   break;
    case STATS_DEBUG3:   g_tmp_debug3_time = tmp_time;   break;
    default:
      printf("ERROR: Unknown timing region\n");
      assert(NULL);   
  }
}







/******************************************************************
  Stats_StopTimer
 ******************************************************************
 
  Description : Gets the end time of a region
  Parameters  : timed_region - integer representing the portion of
                               the solver that was timed
  Modifies    : increments the timer for that region
  Assumptions : StartTimer was called before this function with the
                same timed_region integer
  Returns     : <none>

 ******************************************************************/
void Stats_StopTimer(int timed_region) {
  double end_time = getInternalRunTime();
  switch (timed_region) {
    case STATS_PARSING:  g_parsing_time += (end_time - g_tmp_parsing_time);  break;
    case STATS_MERGING:  g_merging_time += (end_time - g_tmp_merging_time);  break;
    case STATS_COMPACT:  g_compact_time += (end_time - g_tmp_compact_time);  break;
    case STATS_SOLVER:   g_solver_time += (end_time - g_tmp_solver_time);    break;
    case STATS_DEBUG1:   g_debug1_time += (end_time - g_tmp_debug1_time);    break;
    case STATS_DEBUG2:   g_debug2_time += (end_time - g_tmp_debug2_time);    break;
    case STATS_DEBUG3:   g_debug3_time += (end_time - g_tmp_debug3_time);    break;
    default:
      printf("ERROR: Unknown timing region\n");
      assert(NULL);   
  }
}











/************************ Local Functions *************************/
// Local functions were borrowed from 2clseq by Fahiem Bacchus, with
// permission

/*
   GetInternalRunTime

   Description:
   Get the time in seconds since an OS specific epoch.
   If possible, we provide accounted time.
   Notes:
   On non-Windows systems, the default time calculation assumes a Linux, Solaris
   or HPUX system.  Commented out code is also provided for BSD, and if all 
   else fails, there is a call to the ANSI clock routine.  
*/

double getInternalRunTime() {

  double df;
#ifndef WIN32
  // Linux, Solaris, HPUX, etc.
  struct tms usage;
#endif /* win32 */
#ifdef WIN32
  df=getWinRunTime();
#else /* WIN32 */
  times(&usage);
  df=((double)usage.tms_utime+(double)usage.tms_stime)/sysconf(_SC_CLK_TCK);
#endif /* WIN32 */
  return df;
}




#ifdef WIN32
/*
   GetWinRunTime
   
   Description:
   Get the current system time in seconds, as a double precision number.
   For Windows 95/98, this routine reports the elapsed time, using 
   QueryPerformanceCounter and QueryPerformanceFrequency to attain a 
   precision of better than 1 microsecond.
   For Windows NT, this routine reports the accounted CPU time using
   GetProcessTimes.
   Notes:
   Windows 95/98 reports a clock frequency of 1193180 Hz on a PC.  The actual
   frequency is more likely 1193046.471111 Hz, which causes a 32 bit counter
   to overflow exactly once an hour... but we take their word for it anyways.
   We assume that the clock frequency is under 4294967296 Hz, so we can
   ignore the upper longword of the frequency.
*/
static double getWinRunTime() {

  double dfTime;
  if(g_isNT) {
    FILETIME ftTemp;
    FILETIME ftKernal;
    FILETIME ftUser;
  
    GetProcessTimes(GetCurrentProcess(),&ftTemp,&ftTemp,&ftKernal,&ftUser);
    dfTime=(*(__int64 *)&ftKernal+*(__int64 *)&ftUser)/10000000.0;

  } else { /* Windows 95/98 */
    unsigned int n[2];
    static double k0,k1;
  
    if (!k0) {
      if (!QueryPerformanceFrequency((LARGE_INTEGER *)n))
        return 0.0;
      k0=n[0];
      k1=4294967296.0/k0;
    }
    QueryPerformanceCounter((LARGE_INTEGER *)n);
    dfTime=n[1]*k1+n[0]/k0;
  }
  return dfTime;
}
#endif /* WIN32 */


