/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  October 20th 2003 (April 2004)

 ******************************************************************

  Undo.cpp

  This file contains the functions required to undo the assignments
  propagated after a CHOICE.  These functions are highly dependant/
  correlated to the functions in Solver.cpp and NoGood.cpp.  This
  file also contains the required information for restarting once
  a given number of conflicts are encountered.

 ******************************************************************/

#include "undo.h"
#include "stats.h"



/******************* Globally Available Variable ******************/
/*************** Global Variables in this Module ******************/

 // Stack of undo records
int *g_undos = NULL;
int g_num_undos = 0;

 // Current decision level of the DPLL search
int g_decision_level = 0;

 // Restart factors
int g_next_restart = 0;
int g_restart_step = 0;


 // UniqueStack used to remember UIP assignments that are forced
 // from nogoods
UniqueUIPStack *g_uip_assignments;






/****************** Local Function Prototypes *********************/
/*************** Externally Available Functions *******************/




/******************************************************************
  Undo_Initialize
 ******************************************************************
 
  Description : Initialize and allocate memory to the data structs.
  Parameters  : <none>
  Modifies    : allocates memory
  Assumptions : Structures have not already been allocated.
  Returns     : <none>

 ******************************************************************/
void Undo_Initialize() {

 // Initialize the undo stack (this could be made of varying size)
  g_undos = (int*) malloc (g_num_nodes * sizeof(int));
  g_num_undos = 0;
  g_decision_level = 0;
  g_next_restart = SOLVER_RESTART_INITIAL;
  g_restart_step = SOLVER_RESTART_STEP;

 // UIP stack
  g_uip_assignments = new UniqueUIPStack(2 * g_num_nodes);

  assert(g_undos);
  assert(g_uip_assignments);

}






/******************************************************************
  Undo_Free
 ******************************************************************
 
  Description : Deallocates the memory allocated for undos
  Parameters  : <none>
  Modifies    : frees memory
  Assumptions : the structures have already been allocated
  Returns     : <none>

 ******************************************************************/
void Undo_Free() {
  free(g_undos);
  delete g_uip_assignments;
}











/******************************************************************
  Undo_Unroll
 ******************************************************************
 
  Description : Restore all the assignments up to the last decision
                and attempt the other branch.  This function assigns
                all the new assignments discovered as we unroll the
                undo stack.
  Parameters  : node   - the conflict node that forced us to unroll
                reason - the conflict reason that assigned another
                         value to the conflict node
  Modifies    : Removes nodes from the stack, restores their values,
                adds nogoods to the database, determines forced
                assignments and assigns them
  Assumptions : All structures initialized and the first decision
                has level 1
  Returns     : SOLVER_OK or SOLVER_CONFLICT (the result of the
                propagation of all the assignments determined by
                nogoods)

 ******************************************************************/
int Undo_Unroll() {

  int node, retval, old_value = UNKNOWN;
  bool continuing_backtrack = FALSE;
  if (g_args.verbose >= 3) printf("\n\nBegin backtrack\n");


  do {

  // Initialize the dynamic NoGood
    NoGood_CreateNoGood(g_conflict_node, g_conflict_reason, continuing_backtrack);
    continuing_backtrack = TRUE;


   // Unroll the records as far as we can go
    do {

      if (g_decision_level == 1) return SOLVER_UNSATISFIABLE;
      if (g_args.verbose >= 4) printf("\nUnrolling level %d\n", g_decision_level);

      do {

       // Remove each record up to and including the CHOICE
        node = g_undos[--g_num_undos];


       // Replace the node with its reason in the dynamic nogood
        if (g_nodes.value[node] != DONT_CARE)
          NoGood_ExpandUIP(node);


       // Restore the previous value of the node
        old_value = g_nodes.value[node];
        g_nodes.value[node] = UNKNOWN;
        g_nodes.decision_level[node] = 0; // for debugging


      } while ((old_value == DONT_CARE) || 
               (getReasonType(g_nodes.reason[node]) != CHOICE_REASON));


     // Go down to the previous decision level
      g_decision_level--;

    } while (!NoGood_Member(node));


   // Clear the dynamic structure
    NoGood_ResetNoGood();


   // Assign the UIPs
    retval = Undo_SetUIPAssignments();
    if (retval == SOLVER_OK)
      retval = Propagate_Propagate();


  } while (retval != SOLVER_OK);



 // Restart after a given number of conflicts
  if ((g_args.restarts == TRUE) && (g_stats.nogoods_learned >= g_next_restart)) {

   // Remove all the undo records
    if (g_args.verbose >= 2) printf("Restarting...\n");
    node = g_undos[g_num_undos-1];
    while (g_nodes.decision_level[node] != 1) {
      g_nodes.value[node] = UNKNOWN;
      g_nodes.decision_level[node] = 0;  // debugging
      node = g_undos[(--g_num_undos)-1];
    }

   // Fix the stat counters and find the next restart step
    g_decision_level = 1;
    g_stats.num_restarts++;
    g_next_restart += g_restart_step;
    g_restart_step += SOLVER_RESTART_STEP_INCR;

   // Set any UIPs available at level 1
    retval = Undo_SetUIPAssignments();
    if (retval == SOLVER_OK)
      retval = Propagate_Propagate();

    if (retval != SOLVER_OK)
      return SOLVER_UNSATISFIABLE;
  }


  if (g_args.verbose >= 3) printf("End Backtrack\n\n");
  return retval;

}













/******************************************************************
  Undo_CleanupNoGoods
 ******************************************************************
 
  Description : Remove all unnecessary/old nogoods from the nogood
                database
  Parameters  : <none>
  Modifies    : removes nogoods from the database and updates the
                pointers to the nogoods that remain
  Assumptions : quite a few...
  Returns     : <none>
  Note        : ...very messy...

 ******************************************************************/
void Undo_CleanupNoGoods() {

  int i, index, *nogood;
  reasoninfo	reason;


 // Mark all nogoods that are currently "reasons"
  for (i=0; i<g_num_undos; i++) {

    index = g_undos[i];
    reason = g_nodes.reason[index];

    if (g_nodes.value[index] == DONT_CARE)
      continue;


   // Mark all the nogood reasons for any assignments
    if ( ((g_args.nogood_cleanup == TRUE) && (getReasonType(reason) == NOGOOD_REASON))
      || (getReasonType(reason) == UIP_REASON) ) {

     // Get the nogood/uip reason for the assignment and check that
     // it belongs to the temporary database
      nogood = getNoGoodReason(reason);
      assert(nogood);
      if (g_db_permanent_nogoods->member(nogood))
        continue;

     // Mark the nogood so that the cleanup knows it is being used
      g_nodes.reason[index].node = nogood[0];
      nogood[0] = index;
      if (((*(nogood-2)) & 0x80000000) && ((*(nogood-1)) & 0x80000000)) {
        assert(((*(nogood-3)) & 0x3FFFFFFF) == 0);
        *(nogood-3) |= 0x80000000;
      } else {
        assert(((*(nogood-1)) & 0x3FFFFFFF) == 0);
        *(nogood-1) |= 0x80000000;
      }

    }

  }


 // Marks all the UIP nogoods assigned by the previous backtracks
  g_uip_assignments->cleanupUIPs();

}





/******************************************************************
  Undo_VerifyNoGoods
 ******************************************************************
 
  Description : Asserts that the undo records are still pointing
                to the correct nogoods. DEBUGGING FUNCTION!
  Parameters  : <none>
  Modifies    : <none>
  Assumptions : <none>
  Returns     : <none>

 ******************************************************************/
void Undo_VerifyNoGoods() {

  int i, j, index, *nogood;
  reasoninfo	reason;


  for (i=0; i<g_num_undos; i++) {

    index = g_undos[i];
    reason = g_nodes.reason[index];


    if (g_nodes.value[index] == DONT_CARE)
      continue;
      

    if (getReasonType(reason) == NOGOOD_REASON) {

      nogood = getNoGoodReason(reason);

      j=0;
      while (nogood[j] != 0) {
        if ((nogood[j] >> 1) >= g_num_nodes) {
          int k=0;
          printf("Error nogood is:");
          while (nogood[k] != 0) {
            printf(" %d", nogood[k]);
            k++;
          }
          printf("\n");
        }
        assert((nogood[j] >> 1) < g_num_nodes);
        j++;
      }
    }
  }

}







/************************ Local Functions *************************/


