/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  October 20th 2003 (April 2004)

  *****************************************************************

   Undo.h

   Definition of the functions that are needed to remember node
   assignments that can be undone during backtracking.  This file
   contains a UIP stack that is used to determine assignments that
   are learned from the conflict and available after the current
   backtrack.  These functions are inlined for efficiency
   considerations.

  *****************************************************************/

#ifndef __UNDO_H__
#define __UNDO_H__

#include "noclause.h"
#include "uniquestack.h"

#include "solver.h"
#include "nogood.h"
#include "propagate.h"




/************************** Constants *****************************/
/********************** Global Variables **************************/

extern int g_decision_level;
extern int *g_undos;
extern int g_num_undos;





/******************* Structure Declarations ***********************/


// UniqueStack is used to save future assignments.  The stack keeps
// information about the assignment, the nogood reason as well as
// the deepest level to which the assignment applies.

class UniqueUIPStack : public UniqueStack {

protected:

  unsigned long *uip;
  int *level;


public:

 // Constructor
  UniqueUIPStack(int maxsize) : UniqueStack(maxsize) {
    uip = new unsigned long[maxsize];
    level = new int[maxsize];
  }


 // Destructor
  ~UniqueUIPStack() {
    delete[] uip;
    delete[] level;
  }


 // Push a new element to the stack (with UIP ptr and level)
  void inline push(int e, int *u, int l) {

    if ((!marks[e]) || (l < (level[e] & 0x7FFFFFFF))) {
      if (!marks[e]) {
        stack[stack_top++] = e;
        marks[e] = true;
      }
      uip[e] = ((unsigned long)u)  >> 2;
//printf("Pushed %p containing 0x%x into uip[%d] as 0x%lx\n",u,*u,e,uip[e]);
      level[e] = l;
    }

  }



 // Remove the node, UIP and level from the stack
  void inline pop(int *e, int **u, int *l) {
    *e = stack[--stack_top];
    *u = (int *)(uip[*e] << 2);
    *l = level[*e];
    marks[*e] = false;
  }

  int inline getstackentry(int i)
  {
//printf("Get stackentry from uip[%d] (location %p) - return %ld\n",i,&uip[i],uip[i]);
	  return (int) uip[i];
  }

  void inline putstackentry(int i, int *p)
  {
//printf("Putting stackentry at uip[%d] (location %p) to be %p\n",i,&uip[i],p);
	  uip[i] = ((unsigned long)p >> 2);
  }


 // Look through all the UIPs and assign all those applicable at the
 // current decision level, keeping only those that also apply to a
 // lower decision level
  int inline assignUIPs() {

    int i, assign, retval, new_stack_top = 0;

    for (i=0; i<stack_top; i++) {

      assign = stack[i];
      if (level[assign] <= g_decision_level) {
        retval = Solver_Assign(assign >> 1, (assign & 0x1)?TRUE:FALSE,
                               createNoGoodReason(UIP_REASON, (int *) (uip[assign] << 2)));
        assert(retval == SOLVER_OK);
      }

      if (level[assign] < g_decision_level) {
        stack[new_stack_top++] = assign;
      } else {
        marks[assign] = false;
      }

    }

    stack_top = new_stack_top;
    return SOLVER_OK;
  }



 // Mark the UIPs in the database that are used in the UniqueStack
  void inline cleanupUIPs() {

    for (int i=0; i<stack_top; i++) {

      int *ng = (int *)(uip[stack[i]] << 2);
//printf("Dealing with %p uip[%d] stack[%d] containing 0x%x\n",ng,stack[i],i,*ng);
      uip[stack[i]] = ng[0];
//printf("\tuip[%d] (at %p) now equals 0x%lx\n",stack[i],&uip[stack[i]],uip[stack[i]]);
      ng[0] = stack[i];
//printf("\t%p now holds %x(0x%x)\n",ng,stack[i],ng[0]);

      if (((*(ng-2)) & 0x80000000) && ((*(ng-1)) & 0x80000000)) {
        assert(((*(ng-3)) & 0x3FFFFFFF) == 0);
        *(ng-3) |= 0x40000000;
        assert(NULL);
      } else {
        assert(((*(ng-1)) & 0x3FFFFFFF) == 0);
        *(ng-1) |= 0x40000000;
      }

    }

  }



};











/************************* Inline Functions ***********************/

 // This must be here because UIP assignments are used in the inline
 // function in this header
extern UniqueUIPStack *g_uip_assignments;




/******************************************************************
  Undo_Assign
 ******************************************************************
 
  Description : Add the node to the undo stack saving the old value
                and the reason for the assignment
  Parameters  : node - index of the node that is to be saved
                value - new value of the node
                reason - the reason that forced this assignment
  Modifies    : adds an undo record and modifies the value of the
                node, keeping track of extra info such as the
                decision_level
  Assumptions : undos have been initialized
  Returns     : <none>
  Note        : This function is used externally by other files to
                add an undo record to the stack

 ******************************************************************/
void inline Undo_Assign(int node, byte value, reasoninfo reason) {

 // Remember the previous information
  g_undos[g_num_undos++] = node;
  assert(value != UNKNOWN);
  assert(g_nodes.value[node] == UNKNOWN);

 // Increment stat counters (or decision level if a choice node)
  if (value == DONT_CARE) {
    g_stats.num_dc_implications++;
  } else if (getReasonType(reason) != CHOICE_REASON) {
    g_stats.num_implications++;
  } else {
    g_decision_level++;
  }


 // Assign the new value to the node
  g_nodes.value[node] = value;
  g_nodes.reason[node] = reason;
  g_nodes.decision_level[node] = g_decision_level;

}






/******************************************************************
  Undo_AddNoGoodAssignment
 ******************************************************************
 
  Description : Adds a new assignment that has been propagated
                because of a new nogood that we learned
  Parameters  : node  - the node that is to be assigned from the UIP
                uip   - the UIP reason for the assignment
                level - the level down to which this assignment
                        applies (level of 2nd highest node in UIP)
  Modifies    : adds an assignment to the stack
  Assumptions : the decision level of the node is equal to or less
                than the current decision level (else the nogood
                would not be saved)
  Returns     : <none>

 ******************************************************************/
void inline Undo_AddUIPAssignment(int node, int *uip, int level) {

 // Debugging
  assert(level < g_decision_level);

 // Add the assignment, if already there, add it only if the 
 // decision level is lower than the existing assignment
  g_uip_assignments->push(node, uip, level);

}






/******************************************************************
  Undo_SetNoGoodAssignments
 ******************************************************************
 
  Description : Sets all the assignments now that we reached a new
                level.  All assignments that apply to lower decision
                levels are saved, others are discarded.
  Parameters  : <none>
  Modifies    : removes some assignments from the UniqueStack
  Assumptions : everything initialized
  Returns     : SOLVER_OK or SOLVER_CONFLICT depending on the
                results of Solver_Assign

 ******************************************************************/
int inline Undo_SetUIPAssignments() {
  return g_uip_assignments->assignUIPs();
}








/****************** External Function Declarations ****************/

extern void Undo_Initialize();
extern void Undo_Free();
extern int  Undo_Unroll();

 // NoGood Cleanup functions
extern void Undo_CleanupNoGoods();
extern void Undo_VerifyNoGoods();



#endif

