/******************************************************************
  Non-Clausal DPLL Solver
  Version: 2.5

  Christian Thiffault
  University of Toronto
  November 10th 2003 (April 2004)

  *****************************************************************

   UniqueStack.h (based on uniqueset.h in 2clseq by Fahiem Bacchus,
                  with permission)

  Contains the functions required in order to create a constant
  time unique set of integers.

  *****************************************************************/

#ifndef __UNIQUESTACK_H__
#define __UNIQUESTACK_H__





/************************** Constants *****************************/
/******************* Structure Declarations ***********************/


// Stack object where only 1 copy of each integer may be kept
// simultaneously.
class UniqueStack {

protected:

  bool *marks;      // An integer is marked TRUE if present in stack
  int capacity;     // Size of the stack
  int *stack;       // Stack of integers
  int stack_top;    // Last element on the stack


public:

 // Constructor
  UniqueStack(int maxsize) {
    capacity = maxsize;
    stack = new int[capacity];
    stack_top = 0;
    marks = new bool[capacity];
    for (int i=0; i<capacity; i++)
      marks[i] = false;
    assert(stack);
    assert(marks);
  }


 // Destructor  
  ~UniqueStack() {
    delete[] stack;
    delete[] marks;
  }
  

 // Add an element to the stack if not already present  
  void inline push(int e) {
    if (!marks[e]) {
      stack[stack_top++] = e;
      marks[e] = true;
    }
  }
  

 // Remove the top element from the stack  
  int inline pop() {
    marks[ stack[--stack_top] ] = false;
    return stack[stack_top];
  }


 // Check whether the stack has any elements
  bool inline isEmpty() {
    return (stack_top == 0);
  }


 // Check that an element is in the stack
  bool inline member(int e) {
    return marks[e];
  }


 // Remove all the elements from the stack
  void inline empty() {
    for (int i=0; i<stack_top; i++)
      marks[stack[i]] = false;
    stack_top = 0;
  }


 // Get the number of elements currently in the stack
  int inline getSize() {
    return stack_top;
  }

};




// Inherited class that adds a 'void*' information pointer for each
// element in the stack.  The elements are still referenced as ints.
class UniqueInfoStack : public UniqueStack {

protected:

  void **info;       // Array of void pointers


public:

 // Constructor
  UniqueInfoStack(int maxsize) : UniqueStack(maxsize) {
    info = new void*[maxsize];
    assert(info);
  }


 // Destructor  
  ~UniqueInfoStack() {
    delete[] info;
  }


 // Add an element with additional information to the stack
  void inline push(int e, void *i) {

    if (!marks[e]) {
      stack[stack_top++] = e;
      marks[e] = true;      
      info[e] = i;
    }

  }


 // Remove an element from the stack placing the info in the
 // supplied pointer
  int inline pop(void **i) {
    int e = stack[--stack_top];
    marks[e] = false;
    *i = info[e];
    return e;
  }


 // Get the information relating to an element
 //  - Assumption: the element must be in the stack
  void inline *getInfo(int e) {
    return info[e];
  }

};



#endif

